# Filename: fit.py
# -*- coding: utf-8 -*-
# pylint: disable=locally-disabled
"""
A collection of fit functions and modules.

"""
import numpy as np
import km3pipe.extras


def fit_delta_ts(data, time_s):
    """Fits gaussians to delta t for each PMT pair.

    Parameters
    ----------
    data: 2d np.array: x = PMT combinations (465), y = time, entry = frequency
    time_s: length of data taking in seconds

    Returns
    -------
    numpy arrays with rates and means for all PMT combinations
    """
    scipy = km3pipe.extras.scipy()
    from scipy import optimize

    data = data / time_s

    xs = np.arange(-20, 21)

    def gaussian(x, mean, sigma, rate, offset):
        return (
            rate / np.sqrt(2 * np.pi) / sigma * np.exp(-((x - mean) ** 2) / sigma ** 2)
            + offset
        )

    rates = []
    means = []
    for combination in data:
        try:
            popt, _ = optimize.curve_fit(gaussian, xs, combination, p0=[0, 2, 1000, 20])
        except RuntimeError:
            popt = (0, 0, 0, 0)
        rates.append(popt[2])
        means.append(popt[0])
    return np.array(rates), np.array(means)



#!/usr/bin/env python
"""
Example to demonstrate how to modify log levels of KM3Pipe modules.

"""
import km3pipe as kp

core_log = kp.logger.get_logger("km3pipe.core")
core_log.setLevel("DEBUG")


def foo(blob):
    print("Module called")
    return blob


pipe = kp.Pipeline()
pipe.attach(foo)
pipe.drain(5)



# Filename: evt.py
# pylint: disable=C0103,R0903
"""
Pumps for the EVT simulation dataformat.

"""

from collections import defaultdict
import sys

import numpy as np

from thepipe import Module, Blob
from km3pipe.dataclasses import Table
from km3pipe.logger import get_logger
from km3pipe.tools import split

log = get_logger(__name__)  # pylint: disable=C0103

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal, Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def try_decode_string(text):
    """Decode string to ASCII if possible"""
    try:
        return text.decode("ascii")
    except AttributeError:
        return text


class EvtPump(Module):  # pylint: disable:R0902
    """Provides a pump for EVT-files.

    Parameters
    ----------
    filename: str
        The file to read the events from.
    parsers: list of str or callables
        The parsers to apply for each blob (e.g. parsers=['km3sim', a_parser])
        You can also pass your own function, which takes a single argument
        `blob` and mutates it. `str` values will be looked up in the
        `kp.io.evt.EVT_PARSERS` dictionary and ignored if not found.
        If `parsers='auto'`, the `EvtPump` will try to find the appropriate
        parsers, which is the default behaviour. [default: 'auto']
    cache_enabled: bool
        If enabled, a cache of the event indices is created when loading
        the file. Enable it if you want to jump around and inspect the
        events non-consecutively. [default: False]
    basename: str
        The common part of the filenames if you want to process multiple
        files e.g. file1.evt, file2.evt and file3.evt. During processing,
        the files will be concatenated behind the scenes.
        You need to specify the `index_stop` and `index_start`
        (1 and 3 for the example).
    suffix: str
        A string to append to each filename (before ".evt"), when basename
        is given. [default: '']
    index_start: int
        The starting index if you process multiple files at once. [default: 1]
    index_stop: int
        The last index if you process multiple files at once. [default: 1]
    n_digits: int or None
        The number of digits for indexing multiple files. [default: None]
        `None` means no leading zeros.
    exclude_tags: list of strings
        The tags in the EVT file, which should be ignored (e.g. if they
        cause parse errors)

    """

    def configure(self):
        self.filename = self.get("filename", default=None)
        self.filenames = self.get("filenames", default=[])
        parsers = self.get("parsers", default="auto")
        self.cache_enabled = self.get("cache_enabled", default=False)
        self.basename = self.get("basename", default=None)
        self.suffix = self.get("suffix", default="")
        self.index_start = self.get("index_start", default=1)
        if self.filenames:
            self.filename = self.filenames[0]
            self.index_stop = len(self.filenames)
        else:
            self.index_stop = self.get("index_stop", default=1)
        self.n_digits = self.get("n_digits", default=None)
        self.exclude_tags = self.get("exclude_tags")
        if self.exclude_tags is None:
            self.exclude_tags = []

        self.raw_header = None
        self.event_offsets = []
        self.index = 0
        self.whole_file_cached = False
        self.parsers = []
        self._auto_parse = False

        if not self.filename and not self.filenames and not self.basename:
            print("No file- or basename(s) defined!")

        if parsers:
            if parsers == "auto":
                self.cprint("Automatic tag parsing enabled.")
                self._auto_parse = True
            else:
                if isinstance(parsers, str):
                    # expects a list(str)
                    parsers = [parsers]
                self._register_parsers(parsers)

        self.file_index = int(self.index_start)
        if self.filenames:
            self.filename = self.filenames[self.file_index - 1]
        elif self.basename:
            self.log.info(
                "Got a basename ({}), constructing the first "
                "filename.".format(self.basename)
            )
            file_index = self._get_file_index_str()

            self.filename = "{}{}{}.evt".format(self.basename, file_index, self.suffix)
            self.log.info("Constructed filename: {}".format(self.filename))

        if self.filename:
            self.cprint("Opening {0}".format(self.filename))
            self.blob_file = self.open_file(self.filename)
            self.prepare_blobs()

    def _register_parsers(self, parsers):
        self.log.info("Found parsers {}".format(parsers))
        for parser in parsers:
            if callable(parser):
                self.parsers.append(parser)
                continue

            if parser in EVT_PARSERS.keys():
                self.parsers.append(EVT_PARSERS[parser])
            else:
                self.log.warning("Parser '{}' not found, ignoring...".format(parser))

    def _reset(self):
        """Clear the cache."""
        self.log.info("Clearing the cache, resetting event offsets")
        self.raw_header = None
        self.event_offsets = []
        self.index = 0

    def _get_file_index_str(self):
        """Create a string out of the current file_index"""
        file_index = str(self.file_index)
        if self.n_digits is not None:
            file_index = file_index.zfill(self.n_digits)
        return file_index

    def prepare_blobs(self):
        """Populate the blobs"""
        self.raw_header = self.extract_header()
        if self.cache_enabled:
            self._cache_offsets()

    def extract_header(self):
        """Create a dictionary with the EVT header information"""
        self.log.info("Extracting the header")
        raw_header = self.raw_header = defaultdict(list)
        first_line = self.blob_file.readline()
        first_line = try_decode_string(first_line)
        self.blob_file.seek(0, 0)
        if not first_line.startswith(str("start_run")):
            self.log.warning("No header found.")
            return raw_header
        for line in iter(self.blob_file.readline, ""):
            line = try_decode_string(line)
            line = line.strip()
            try:
                tag, value = str(line).split(":")
            except ValueError:
                continue
            raw_header[tag].append(str(value).split())
            if line.startswith(str("end_event:")):
                self._record_offset()
                if self._auto_parse and "physics" in raw_header:
                    parsers = [p[0].lower() for p in raw_header["physics"]]
                    self._register_parsers(parsers)
                return raw_header
        raise ValueError("Incomplete header, no 'end_event' tag found!")

    def get_blob(self, index):
        """Return a blob with the event at the given index"""
        self.log.info("Retrieving blob #{}".format(index))
        if index > len(self.event_offsets) - 1:
            self.log.info("Index not in cache, caching offsets")
            self._cache_offsets(index, verbose=False)
        self.blob_file.seek(self.event_offsets[index], 0)
        blob = self._create_blob()
        if blob is None:
            self.log.info("Empty blob created...")
            raise IndexError
        else:
            self.log.debug("Applying parsers...")
            for parser in self.parsers:
                parser(blob)
            self.log.debug("Returning the blob")
            return blob

    def process(self, blob=None):
        """Pump the next blob to the modules"""
        try:
            blob = self.get_blob(self.index)

        except IndexError:
            self.log.info("Got an IndexError, trying the next file")
            if (self.basename or self.filenames) and self.file_index < self.index_stop:
                self.file_index += 1
                self.log.info("Now at file_index={}".format(self.file_index))
                self._reset()
                self.blob_file.close()
                self.log.info("Resetting blob index to 0")
                self.index = 0
                file_index = self._get_file_index_str()
                if self.filenames:
                    self.filename = self.filenames[self.file_index - 1]
                elif self.basename:
                    self.filename = "{}{}{}.evt".format(
                        self.basename, file_index, self.suffix
                    )
                self.log.info("Next filename: {}".format(self.filename))
                self.cprint("Opening {0}".format(self.filename))
                self.blob_file = self.open_file(self.filename)
                self.prepare_blobs()
                try:
                    blob = self.get_blob(self.index)
                except IndexError:
                    self.log.warning("No blob found in file {}".format(self.filename))
                else:
                    return blob
            self.log.info("No files left, terminating the pipeline")
            raise StopIteration

        self.index += 1
        return blob

    def _cache_offsets(self, up_to_index=None, verbose=True):
        """Cache all event offsets."""
        if not up_to_index:
            if verbose:
                self.cprint("Caching event file offsets, this may take a bit.")
            self.blob_file.seek(0, 0)
            self.event_offsets = []
            if not self.raw_header:
                self.event_offsets.append(0)
        else:
            self.blob_file.seek(self.event_offsets[-1], 0)
        for line in iter(self.blob_file.readline, ""):
            line = try_decode_string(line)
            if line.startswith("end_event:"):
                self._record_offset()
                if len(self.event_offsets) % 100 == 0:
                    if verbose:
                        print(".", end="")
                    sys.stdout.flush()
            if up_to_index and len(self.event_offsets) >= up_to_index + 1:
                return
        self.event_offsets.pop()  # get rid of the last entry
        if not up_to_index:
            self.whole_file_cached = True
        self.cprint("\n{0} events indexed.".format(len(self.event_offsets)))

    def _record_offset(self):
        """Stores the current file pointer position"""
        offset = self.blob_file.tell()
        self.event_offsets.append(offset)

    def _create_blob(self):
        """Parse the next event from the current file position"""
        blob = None
        for line in self.blob_file:
            line = try_decode_string(line)
            line = line.strip()
            if line == "":
                self.log.info("Ignoring empty line...")
                continue
            if line.startswith("end_event:") and blob:
                blob["raw_header"] = self.raw_header
                return blob
            try:
                tag, values = line.split(":")
            except ValueError:
                self.log.warning("Ignoring corrupt line: {}".format(line))
                continue
            try:
                values = tuple(split(values.strip(), callback=float))
            except ValueError:
                self.log.info("Empty value: {}".format(values))
            if line.startswith("start_event:"):
                blob = Blob()
                blob[tag] = tuple(int(v) for v in values)
                continue
            if tag not in blob:
                blob[tag] = []
            blob[tag].append(values)

    def __len__(self):
        if not self.whole_file_cached:
            self._cache_offsets()
        return len(self.event_offsets)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            blob = self.get_blob(self.index)
        except IndexError:
            self.index = 0
            raise StopIteration
        self.index += 1
        return blob

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.get_blob(index)
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield self.get_blob(i)

    def finish(self):
        if self.blob_file:
            self.blob_file.close()


class Parser(object):
    """Standard parser to create numpy times from EVT raw data.

    The `tag_description` is a dict of tuples. The key is the target blob-key,
    the value is tuple of "target blob-key" and "numpy dtype".

    """

    def __init__(self, tag_description):
        self.tag_description = tag_description

    def __call__(self, blob):
        """Iterate through the blob-keys and add the parsed data to the blob"""
        for key in list(blob.keys()):
            if key in self.tag_description.keys():
                data = blob[key]
                out_key, dtype = self.tag_description[key]
                arr = np.array([d[: len(dtype)] for d in data], dtype)
                tab = Table(arr, name=out_key)
                blob[out_key] = tab


KM3SIM_TAGS = {
    "hit": [
        "KM3SimHits",
        [
            ("id", "f4"),
            ("pmt_id", "<i4"),
            ("pe", "f4"),
            ("time", "f4"),
            ("type", "f4"),
            ("n_photons", "f4"),
            ("track_in", "f4"),
            ("c_time", "f4"),
            ("unknown", "f4"),
        ],
    ],
}

GSEAGEN_TAGS = {
    "neutrino": [
        "Neutrinos",
        [
            ("id", "<i4"),
            ("pos_x", "f4"),
            ("pos_y", "f4"),
            ("pos_z", "f4"),
            ("dir_x", "f4"),
            ("dir_y", "f4"),
            ("dir_z", "f4"),
            ("energy", "f4"),
            ("time", "f4"),
            ("bjorken_x", "f4"),
            ("bjorken_y", "f4"),
            ("scattering_type", "<i4"),
            ("pdg_id", "<i4"),
            ("interaction_type", "<i4"),
        ],
    ],
    "track_in": [
        "TrackIns",
        [
            ("id", "<i4"),
            ("pos_x", "f4"),
            ("pos_y", "f4"),
            ("pos_z", "f4"),
            ("dir_x", "f4"),
            ("dir_y", "f4"),
            ("dir_z", "f4"),
            ("energy", "f4"),
            ("time", "f4"),
            ("geant_id", "f4"),
        ],
    ],
    "primary_lepton": [
        "PrimaryLeptons",
        [
            ("id", "<i4"),
            ("pos_x", "f4"),
            ("pos_y", "f4"),
            ("pos_z", "f4"),
            ("dir_x", "f4"),
            ("dir_y", "f4"),
            ("dir_z", "f4"),
            ("energy", "f4"),
            ("time", "f4"),
            ("geant_id", "f4"),
        ],
    ],
}

KM3_TAGS = {
    "neutrino": [
        "Neutrinos",
        [
            ("id", "<i4"),
            ("pos_x", "f4"),
            ("pos_y", "f4"),
            ("pos_z", "f4"),
            ("dir_x", "f4"),
            ("dir_y", "f4"),
            ("dir_z", "f4"),
            ("energy", "f4"),
            ("time", "f4"),
            ("bjorken_x", "f4"),
            ("bjorken_y", "f4"),
            ("scattering_type", "<i4"),
            ("pdg_id", "<i4"),
            ("interaction_type", "<i4"),
        ],
    ],
    "track_in": [
        "TrackIns",
        [
            ("id", "<i4"),
            ("pos_x", "f4"),
            ("pos_y", "f4"),
            ("pos_z", "f4"),
            ("dir_x", "f4"),
            ("dir_y", "f4"),
            ("dir_z", "f4"),
            ("energy", "f4"),
            ("time", "f4"),
            ("type", "f4"),
            ("something", "<i4"),
        ],
    ],
    "hit_raw": [
        "Hits",
        [
            ("id", "<i4"),
            ("pmt_id", "<i4"),
            ("npe", "<i4"),
            ("time", "f4"),
        ],
    ],
}


def parse_corant(blob):
    """Creates new blob entries for the given blob keys"""

    if "track_seamuon" in blob.keys():

        muon = blob["track_seamuon"]

        blob["Muon"] = Table(
            {
                "id": np.array(muon)[:, 0].astype(int),
                "pos_x": np.array(muon)[:, 1],
                "pos_y": np.array(muon)[:, 2],
                "pos_z": np.array(muon)[:, 3],
                "dir_x": np.array(muon)[:, 4],
                "dir_y": np.array(muon)[:, 5],
                "dir_z": np.array(muon)[:, 6],
                "energy": np.array(muon)[:, 7],
                "time": np.array(muon)[:, 8],
                "particle_id": np.array(muon)[:, 9].astype(int),
                "is_charm": np.array(muon)[:, 10].astype(int),
                "mother_pid": np.array(muon)[:, 11].astype(int),
                "grandmother_pid": np.array(muon)[:, 11].astype(int),
            },
            h5loc="muon",
        )

        blob["MuonMultiplicity"] = Table(
            {"muon_multiplicity": len(np.array(muon)[:, 6])}, h5loc="muon_multiplicity"
        )

    if "track_seaneutrino" in blob.keys():

        nu = blob["track_seaneutrino"]

        blob["Neutrino"] = Table(
            {
                "id": np.array(nu)[:, 0].astype(int),
                "pos_x": np.array(nu)[:, 1],
                "pos_y": np.array(nu)[:, 2],
                "pos_z": np.array(nu)[:, 3],
                "dir_x": np.array(nu)[:, 4],
                "dir_y": np.array(nu)[:, 5],
                "dir_z": np.array(nu)[:, 6],
                "energy": np.array(nu)[:, 7],
                "time": np.array(nu)[:, 8],
                "particle_id": np.array(nu)[:, 9].astype(int),
                "is_charm": np.array(nu)[:, 10].astype(int),
                "mother_pid": np.array(nu)[:, 11].astype(int),
                "grandmother_pid": np.array(nu)[:, 11].astype(int),
            },
            h5loc="nu",
        )
        blob["NeutrinoMultiplicity"] = Table(
            {
                "total": len(np.array(nu)[:, 6]),
                "nue": len(np.array(nu)[:, 6][np.array(nu)[:, 9] == 66]),
                "anue": len(np.array(nu)[:, 6][np.array(nu)[:, 9] == 67]),
                "numu": len(np.array(nu)[:, 6][np.array(nu)[:, 9] == 68]),
                "anumu": len(np.array(nu)[:, 6][np.array(nu)[:, 9] == 69]),
            },
            h5loc="nu_multiplicity",
        )

    if ("track_seamuon" or "track_seaneutrino") in blob.keys():

        blob["Weights"] = Table(
            {
                "w1": blob["weights"][0][0],
                "w2": blob["weights"][0][1],
                "w3": blob["weights"][0][2],
            },
            h5loc="weights",
        )

    if "track_primary" in blob.keys():

        primary = blob["track_primary"]

        blob["Primary"] = Table(
            {
                "id": np.array(primary)[:, 0].astype(int),
                "pos_x": np.array(primary)[:, 1],
                "pos_y": np.array(primary)[:, 2],
                "pos_z": np.array(primary)[:, 3],
                "dir_x": np.array(primary)[:, 4],
                "dir_y": np.array(primary)[:, 5],
                "dir_z": np.array(primary)[:, 6],
                "energy": np.array(primary)[:, 7],
                "time": np.array(primary)[:, 8],
                "particle_id": np.array(primary)[:, 9].astype(int),
            },
            h5loc="primary",
        )

    return blob


def parse_propa(blob):
    """Creates new blob entries for the given blob keys"""

    if "track_in" in blob.keys():

        muon = blob["track_in"]

        blob["Muon"] = Table(
            {
                "id": np.array(muon)[:, 0].astype(int),
                "pos_x": np.array(muon)[:, 1],
                "pos_y": np.array(muon)[:, 2],
                "pos_z": np.array(muon)[:, 3],
                "dir_x": np.array(muon)[:, 4],
                "dir_y": np.array(muon)[:, 5],
                "dir_z": np.array(muon)[:, 6],
                "energy": np.array(muon)[:, 7],
                "time": np.array(muon)[:, 8],
                "particle_id": np.array(muon)[:, 9].astype(int),
                "is_charm": np.array(muon)[:, 10].astype(int),
                "mother_pid": np.array(muon)[:, 11].astype(int),
                "grandmother_pid": np.array(muon)[:, 11].astype(int),
            },
            h5loc="muon",
        )

        blob["MuonMultiplicity"] = Table(
            {"muon_multiplicity": len(np.array(muon)[:, 6])}, h5loc="muon_multiplicity"
        )

    if "neutrino" in blob.keys():

        nu = blob["neutrino"]

        blob["Neutrino"] = Table(
            {
                "id": np.array(nu)[:, 0].astype(int),
                "pos_x": np.array(nu)[:, 1],
                "pos_y": np.array(nu)[:, 2],
                "pos_z": np.array(nu)[:, 3],
                "dir_x": np.array(nu)[:, 4],
                "dir_y": np.array(nu)[:, 5],
                "dir_z": np.array(nu)[:, 6],
                "energy": np.array(nu)[:, 7],
                "time": np.array(nu)[:, 8],
                "particle_id": np.array(nu)[:, 9].astype(int),
                "is_charm": np.array(nu)[:, 10].astype(int),
                "mother_pid": np.array(nu)[:, 11].astype(int),
                "grandmother_pid": np.array(nu)[:, 11].astype(int),
            },
            h5loc="nu",
        )
        blob["NeutrinoMultiplicity"] = Table(
            {
                "total": len(np.array(nu)[:, 6]),
                "nue": len(np.array(nu)[:, 6][np.array(nu)[:, 9] == 12]),
                "anue": len(np.array(nu)[:, 6][np.array(nu)[:, 9] == -12]),
                "numu": len(np.array(nu)[:, 6][np.array(nu)[:, 9] == 14]),
                "anumu": len(np.array(nu)[:, 6][np.array(nu)[:, 9] == -14]),
            },
            h5loc="nu_multiplicity",
        )

    if ("track_in" or "neutrino") in blob.keys():

        blob["Weights"] = Table(
            {
                "w1": blob["weights"][0][0],
                "w2": blob["weights"][0][1],
                "w3": blob["weights"][0][2],
            },
            h5loc="weights",
        )

    if "track_primary" in blob.keys():

        primary = blob["track_primary"]

        blob["Primary"] = Table(
            {
                "id": np.array(primary)[:, 0].astype(int),
                "pos_x": np.array(primary)[:, 1],
                "pos_y": np.array(primary)[:, 2],
                "pos_z": np.array(primary)[:, 3],
                "dir_x": np.array(primary)[:, 4],
                "dir_y": np.array(primary)[:, 5],
                "dir_z": np.array(primary)[:, 6],
                "energy": np.array(primary)[:, 7],
                "time": np.array(primary)[:, 8],
                "particle_id": np.array(primary)[:, 9].astype(int),
            },
            h5loc="primary",
        )

    return blob


EVT_PARSERS = {
    "km3sim": Parser(KM3SIM_TAGS),
    "gseagen": Parser(GSEAGEN_TAGS),
    "km3": Parser(KM3_TAGS),
    "propa": parse_propa,
    "corant": parse_corant,
}



import curses


class PipeInspector(object):
    def __init__(self, stdscr):
        print("Initialising PipeInspector")
        self.stdscr = stdscr
        self.max_row = None
        self.max_col = None
        self.origin = (0, 0)

        self._update_dimensions()

        self._windows = []
        self._pads = []

        self._setup_curses()
        self._create_windows()
        # self._create_pads()
        self.refresh()

    def test(self):
        self.stdscr.addstr("Pretty text", curses.color_pair(1))

    def run(self):
        self.refresh()
        while True:
            try:
                c = self.stdscr.getkey()
                if c == "q":
                    raise SystemExit
            except curses.error:
                self.refresh()

    def refresh(self):
        # resize = curses.is_term_resized(self.max_row, self.max_col)
        # if resize:
        #    self._update_dimensions()
        #    self.stdscr.clear()
        #    curses.resizeterm(self.max_row, self.max_col)
        for pad in self._pads:
            pad.refresh(0, 0, 1, 1, self.max_row - 1, self.max_col - 1)
        for window in self._windows:
            window.box()
            window.refresh()
        self.stdscr.refresh()

    def _setup_curses(self):
        curses.noecho()
        curses.cbreak()
        self.stdscr.keypad(True)
        curses.init_pair(1, curses.COLOR_RED, curses.COLOR_WHITE)
        self.stdscr.border(0)
        self.stdscr.immedok(True)

    def _update_dimensions(self):
        self.max_row, self.max_col = self.stdscr.getmaxyx()
        self.max_row -= 1
        self.max_col -= 1

    def _create_windows(self):
        x = 20
        y = 7
        height = 5
        width = 40
        win = curses.newwin(height, width, y, x)
        win.box()
        win.immedok(True)
        self._windows.append(win)

    def _create_pads(self):
        pad = curses.newpad(100, 100)
        pad.box()
        pad.immedok(True)
        for y in range(0, 100):
            for x in range(0, 100):
                try:
                    pad.addch(y, x, ord("a") + (x * x + y * y) % 26)
                except curses.error:
                    pass
        self._pads.append(pad)


def main(stdscr):
    stdscr.clear()
    pipe_inspector = PipeInspector(stdscr)
    pipe_inspector.run()


if __name__ == "__main__":
    curses.wrapper(main)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
=============
Ligier Mirror
=============

Subscribes to given tag(s) and sends them to another Ligier.
This script is also available as a command line utility in km3pipe, which can
be accessed by the command ``ligiermirror``.

"""

# Author: Tamas Gal <tgal@km3net.de>
# License: MIT

import socket

from km3pipe import Pipeline, Module
from km3pipe.io import CHPump


class LigierSender(Module):
    def configure(self):
        self.ligier = self.get("ligier") or "127.0.0.1"
        self.port = self.get("port") or 5553
        self.socket = socket.socket()
        self.client = self.socket.connect((self.ligier, self.port))

    def process(self, blob):
        self.socket.send(blob["CHPrefix"].data + blob["CHData"])

    def finish(self):
        self.socket.close()


pipe = Pipeline()
pipe.attach(
    CHPump,
    host="192.168.0.121",
    port=5553,
    tags="IO_EVT, IO_SUM, IO_TSL",
    timeout=60 * 60 * 24 * 7,
    max_queue=2000,
)
pipe.attach(LigierSender)
pipe.drain()



# Filename: hdf5.py
# pylint: disable=C0103,R0903,C901
# vim:set ts=4 sts=4 sw=4 et:
"""
Read and write KM3NeT-formatted HDF5 files.

"""

from collections import OrderedDict, defaultdict, namedtuple
from functools import singledispatch
import os.path
import warnings
from uuid import uuid4

import numpy as np
import tables as tb
import km3io
from thepipe import Provenance

try:
    from numba import jit
except ImportError:
    jit = lambda f: f

import km3pipe as kp
from thepipe import Module, Blob
from km3pipe.dataclasses import Table, NDArray
from km3pipe.logger import get_logger
from km3pipe.tools import decamelise, camelise, split, istype

log = get_logger(__name__)  # pylint: disable=C0103

__author__ = "Tamas Gal and Moritz Lotze and Michael Moser"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"

FORMAT_VERSION = np.string_("5.1")
MINIMUM_FORMAT_VERSION = np.string_("4.1")


class H5VersionError(Exception):
    pass


def check_version(h5file):
    try:
        version = np.string_(h5file.root._v_attrs.format_version)
    except AttributeError:
        log.error(
            "Could not determine HDF5 format version: '%s'."
            "You may encounter unexpected errors! Good luck..." % h5file.filename
        )
        return
    if split(version, int, np.string_(".")) < split(
        MINIMUM_FORMAT_VERSION, int, np.string_(".")
    ):
        raise H5VersionError(
            "HDF5 format version {0} or newer required!\n"
            "'{1}' has HDF5 format version {2}.".format(
                MINIMUM_FORMAT_VERSION.decode("utf-8"),
                h5file.filename,
                version.decode("utf-8"),
            )
        )


class HDF5Header(object):
    """Wrapper class for the `/raw_header` table in KM3HDF5

    Parameters
    ----------
    data : dict(str, str/tuple/dict/OrderedDict)
      The actual header data, consisting of a key and an entry.
      If possible, the key will be set as a property and the the values will
      be converted to namedtuples (fields sorted by name to ensure consistency
      when dictionaries are provided).

    """

    def __init__(self, data):
        self._data = data
        self._user_friendly_data = {}  # namedtuples, if possible
        self._set_attributes()

    def _set_attributes(self):
        """Traverse the internal dictionary and set the getters"""
        for parameter in list(self._data.keys()):
            data = self._data[parameter]
            if isinstance(data, dict) or isinstance(data, OrderedDict):
                if not all(f.isidentifier() for f in data.keys()):
                    break
                # Create a namedtuple for easier access
                field_names, field_values = zip(*data.items())
                sorted_indices = np.argsort(field_names)
                clsname = "HeaderEntry" if not parameter.isidentifier() else parameter
                nt = namedtuple(clsname, [field_names[i] for i in sorted_indices])
                data = nt(*[field_values[i] for i in sorted_indices])
            if parameter.isidentifier():
                setattr(self, parameter, data)
            self._user_friendly_data[parameter] = data

    def __getitem__(self, key):
        return self._user_friendly_data[key]

    def keys(self):
        return self._user_friendly_data.keys()

    def values(self):
        return self._user_friendly_data.values()

    def items(self):
        return self._user_friendly_data.items()

    @classmethod
    def from_table(cls, table):
        data = OrderedDict()
        for i in range(len(table)):
            parameter = table["parameter"][i].decode()
            field_names = table["field_names"][i].decode().split(" ")
            field_values = table["field_values"][i].decode().split(" ")
            if field_values == [""]:
                log.info("No value for parameter '{}'! Skipping...".format(parameter))
                continue
            dtypes = table["dtype"][i].decode()
            dtyped_values = []
            for dtype, value in zip(dtypes.split(" "), field_values):
                if dtype.startswith("a"):
                    dtyped_values.append(value)
                else:
                    value = np.fromstring(value, dtype=dtype, sep=" ")[0]
                    dtyped_values.append(value)
            data[parameter] = OrderedDict(zip(field_names, dtyped_values))
        return cls(data)

    @classmethod
    def from_km3io(cls, header):
        if not isinstance(header, km3io.offline.Header):
            raise TypeError(
                "The given header object is not an instance of km3io.offline.Header"
            )
        return cls(header._data)

    @classmethod
    def from_aanet(cls, table):
        data = OrderedDict()
        for i in range(len(table)):
            parameter = table["parameter"][i].astype(str)
            field_names = [n.decode() for n in table["field_names"][i].split()]
            field_values = [n.decode() for n in table["field_values"][i].split()]
            if field_values in [[b""], []]:
                log.info("No value for parameter '{}'! Skipping...".format(parameter))
                continue
            dtypes = table["dtype"][i]
            dtyped_values = []
            for dtype, value in zip(dtypes.split(), field_values):
                if dtype.startswith(b"a"):
                    dtyped_values.append(value)
                else:
                    value = np.fromstring(value, dtype=dtype, sep=" ")[0]
                    dtyped_values.append(value)
            data[parameter] = OrderedDict(zip(field_names, dtyped_values))
        return cls(data)

    @classmethod
    def from_hdf5(cls, filename):
        with tb.open_file(filename, "r") as f:
            table = f.get_node("/raw_header")
            return cls.from_pytable(table)

    @classmethod
    def from_pytable(cls, table):
        data = OrderedDict()
        for row in table:
            parameter = row["parameter"].decode()
            field_names = row["field_names"].decode().split(" ")
            field_values = row["field_values"].decode().split(" ")
            if field_values == [""]:
                log.info("No value for parameter '{}'! Skipping...".format(parameter))
                continue
            dtypes = row["dtype"].decode()
            dtyped_values = []
            for dtype, value in zip(dtypes.split(" "), field_values):
                if dtype.startswith("a"):
                    dtyped_values.append(value)
                else:
                    value = np.fromstring(value, dtype=dtype, sep=" ")[0]
                    dtyped_values.append(value)
            data[parameter] = OrderedDict(zip(field_names, dtyped_values))
        return cls(data)


class HDF5IndexTable:
    def __init__(self, h5loc, start=0):
        self.h5loc = h5loc
        self._data = defaultdict(list)
        self._index = 0
        if start > 0:
            self._data["indices"] = [0] * start
            self._data["n_items"] = [0] * start

    def append(self, n_items):
        self._data["indices"].append(self._index)
        self._data["n_items"].append(n_items)
        self._index += n_items

    @property
    def data(self):
        return self._data

    def fillup(self, length):
        missing = length - len(self)
        self._data["indices"] += [self.data["indices"][-1]] * missing
        self._data["n_items"] += [0] * missing

    def __len__(self):
        return len(self.data["indices"])


class HDF5Sink(Module):
    """Write KM3NeT-formatted HDF5 files, event-by-event.

    The data can be a ``kp.Table``, a numpy structured array,
    a pandas DataFrame, or a simple scalar.

    The name of the corresponding H5 table is the decamelised
    blob-key, so values which are stored in the blob under `FooBar`
    will be written to `/foo_bar` in the HDF5 file.

    Parameters
    ----------
    filename: str, optional [default: 'dump.h5']
        Where to store the events.
    h5file: pytables.File instance, optional [default: None]
        Opened file to write to. This is mutually exclusive with filename.
    keys: list of strings, optional
        List of Blob-keys to write, everything else is ignored.
    complib : str [default: zlib]
        Compression library that should be used.
        'zlib', 'lzf', 'blosc' and all other PyTables filters
        are available.
    complevel : int [default: 5]
        Compression level.
    chunksize : int [optional]
        Chunksize that should be used for saving along the first axis
        of the input array.
    flush_frequency: int, optional [default: 500]
        The number of iterations to cache tables and arrays before
        dumping to disk.
    pytab_file_args: dict [optional]
        pass more arguments to the pytables File init
    n_rows_expected = int, optional [default: 10000]
    append: bool, optional [default: False]
    reset_group_id: bool, optional [default: True]
        Resets the group_id so that it's continuous in the output file.
        Use this with care!

    Notes
    -----
    Provides service write_table(tab, h5loc=None): tab:Table, h5loc:str
        The table to write, with ".h5loc" set or to h5loc if specified.

    """

    def configure(self):
        self.filename = self.get("filename", default="dump.h5")
        self.ext_h5file = self.get("h5file")
        self.keys = self.get("keys", default=[])
        self.complib = self.get("complib", default="zlib")
        self.complevel = self.get("complevel", default=5)
        self.chunksize = self.get("chunksize")
        self.flush_frequency = self.get("flush_frequency", default=500)
        self.pytab_file_args = self.get("pytab_file_args", default=dict())
        self.keep_open = self.get("keep_open")
        self._reset_group_id = self.get("reset_group_id", default=True)
        self.indices = {}  # to store HDF5IndexTables for each h5loc
        self._singletons_written = {}
        # magic 10000: this is the default of the "expectedrows" arg
        # from the tables.File.create_table() function
        # at least according to the docs
        # might be able to set to `None`, I don't know...
        self.n_rows_expected = self.get("n_rows_expected", default=10000)
        self.index = 0
        self._uuid = str(uuid4())

        self.expose(self.write_table, "write_table")

        if self.ext_h5file is not None:
            self.h5file = self.ext_h5file
        else:
            self.h5file = tb.open_file(
                self.filename,
                mode="w",
                title="KM3NeT",
                **self.pytab_file_args,
            )
            Provenance().record_output(
                self.filename, uuid=self._uuid, comment="HDF5Sink output"
            )
        self.filters = tb.Filters(
            complevel=self.complevel,
            shuffle=True,
            fletcher32=True,
            complib=self.complib,
        )
        self._tables = OrderedDict()
        self._ndarrays = OrderedDict()
        self._ndarrays_cache = defaultdict(list)

    def _to_array(self, data, name=None):
        if data is None:
            return
        if np.isscalar(data):
            self.log.debug("toarray: is a scalar")
            return Table(
                {name: np.asarray(data).reshape((1,))},
                h5loc="/misc/{}".format(decamelise(name)),
                name=name,
            )
        if hasattr(data, "len") and len(data) <= 0:  # a bit smelly ;)
            self.log.debug("toarray: data has no length")
            return
        # istype instead isinstance, to avoid heavy pandas import (hmmm...)
        if istype(data, "DataFrame"):  # noqa
            self.log.debug("toarray: pandas dataframe")
            data = Table.from_dataframe(data)
        return data

    def _cache_ndarray(self, arr):
        self._ndarrays_cache[arr.h5loc].append(arr)

    def _write_ndarrays_cache_to_disk(self):
        """Writes all the cached NDArrays to disk and empties the cache"""
        for h5loc, arrs in self._ndarrays_cache.items():
            title = arrs[0].title
            chunkshape = (
                (self.chunksize,) + arrs[0].shape[1:]
                if self.chunksize is not None
                else None
            )

            arr = NDArray(np.concatenate(arrs), h5loc=h5loc, title=title)

            if h5loc not in self._ndarrays:
                loc, tabname = os.path.split(h5loc)
                ndarr = self.h5file.create_earray(
                    loc,
                    tabname,
                    tb.Atom.from_dtype(arr.dtype),
                    (0,) + arr.shape[1:],
                    chunkshape=chunkshape,
                    title=title,
                    filters=self.filters,
                    createparents=True,
                )
                self._ndarrays[h5loc] = ndarr
            else:
                ndarr = self._ndarrays[h5loc]

            # for arr_length in (len(a) for a in arrs):
            #     self._record_index(h5loc, arr_length)

            ndarr.append(arr)

        self._ndarrays_cache = defaultdict(list)

    def write_table(self, table, h5loc=None):
        """Write a single table to the HDF5 file, exposed as a service"""
        self.log.debug("Writing table %s", table.name)
        if h5loc is None:
            h5loc = table.h5loc
        self._write_table(h5loc, table, table.name)

    def _write_table(self, h5loc, arr, title):
        level = len(h5loc.split("/"))

        if h5loc not in self._tables:
            dtype = arr.dtype
            if any("U" in str(dtype.fields[f][0]) for f in dtype.fields):
                self.log.error(
                    "Cannot write data to '{}'. Unicode strings are not supported!".format(
                        h5loc
                    )
                )
                return
            loc, tabname = os.path.split(h5loc)
            self.log.debug(
                "h5loc '{}', Loc '{}', tabname '{}'".format(h5loc, loc, tabname)
            )
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", tb.NaturalNameWarning)
                tab = self.h5file.create_table(
                    loc,
                    tabname,
                    chunkshape=self.chunksize,
                    description=dtype,
                    title=title,
                    filters=self.filters,
                    createparents=True,
                    expectedrows=self.n_rows_expected,
                )
            tab._v_attrs.datatype = title
            if level < 5:
                self._tables[h5loc] = tab
        else:
            tab = self._tables[h5loc]

        h5_colnames = set(tab.colnames)
        tab_colnames = set(arr.dtype.names)
        if h5_colnames != tab_colnames:
            missing_cols = h5_colnames - tab_colnames
            if missing_cols:
                self.log.info("Missing columns in table, trying to append NaNs.")
                arr = arr.append_columns(
                    missing_cols, np.full((len(missing_cols), len(arr)), np.nan)
                )
                if arr.dtype != tab.dtype:
                    self.log.error(
                        "Differing dtypes after appending "
                        "missing columns to the table! Skipping..."
                    )
                    return

        if arr.dtype != tab.dtype:
            try:
                arr = Table(arr, dtype=tab.dtype)
            except ValueError:
                self.log.critical(
                    "Cannot write a table to '%s' since its dtype is "
                    "different compared to the previous table with the same "
                    "HDF5 location, which was used to fix the dtype of the "
                    "HDF5 compund type." % h5loc
                )
                raise

        tab.append(arr)

        if level < 4:
            tab.flush()

    def _write_separate_columns(self, where, obj, title):
        f = self.h5file
        loc, group_name = os.path.split(where)
        if where not in f:
            group = f.create_group(loc, group_name, title, createparents=True)
            group._v_attrs.datatype = title
        else:
            group = f.get_node(where)

        for col, (dt, _) in obj.dtype.fields.items():
            data = obj.__array__()[col]

            if col not in group:
                a = tb.Atom.from_dtype(dt)
                arr = f.create_earray(
                    group, col, a, (0,), col.capitalize(), filters=self.filters
                )
            else:
                arr = getattr(group, col)
            arr.append(data)

        # create index table
        # if where not in self.indices:
        #     self.indices[where] = HDF5IndexTable(where + "/_indices", start=self.index)

        self._record_index(where, len(data), split=True)

    def _process_entry(self, key, entry):
        self.log.debug("Inspecting {}".format(key))
        if (
            hasattr(entry, "h5singleton")
            and entry.h5singleton
            and entry.h5loc in self._singletons_written
        ):
            self.log.debug(
                "Skipping '%s' since it's a singleton and already written."
                % entry.h5loc
            )
            return
        if not hasattr(entry, "h5loc"):
            self.log.debug("Ignoring '%s': no h5loc attribute" % key)
            return

        if isinstance(entry, NDArray):
            self._cache_ndarray(entry)
            self._record_index(entry.h5loc, len(entry))
            return entry
        try:
            title = entry.name
        except AttributeError:
            title = key

        if isinstance(entry, Table) and not entry.h5singleton:
            if "group_id" not in entry:
                entry = entry.append_columns("group_id", self.index)
            elif self._reset_group_id:
                # reset group_id to the HDF5Sink's continuous counter
                entry.group_id = self.index

        self.log.debug("h5l: '{}', title '{}'".format(entry.h5loc, title))

        if hasattr(entry, "split_h5") and entry.split_h5:
            self.log.debug("Writing into separate columns...")
            self._write_separate_columns(entry.h5loc, entry, title=title)
        else:
            self.log.debug("Writing into single Table...")
            self._write_table(entry.h5loc, entry, title=title)

        if hasattr(entry, "h5singleton") and entry.h5singleton:
            self._singletons_written[entry.h5loc] = True

        return entry

    def process(self, blob):
        written_blob = Blob()
        for key, entry in sorted(blob.items()):
            if self.keys and key not in self.keys:
                self.log.info("Skipping blob, since it's not in the keys list")
                continue
            self.log.debug("Processing %s", key)
            data = self._process_entry(key, entry)
            if data is not None:
                written_blob[key] = data

        if "GroupInfo" not in blob:
            gi = Table(
                {"group_id": self.index, "blob_length": len(written_blob)},
                h5loc="/group_info",
                name="Group Info",
            )
            self._process_entry("GroupInfo", gi)

        # fill up NDArray indices with 0 entries if needed
        if written_blob:
            ndarray_h5locs = set(self._ndarrays.keys()).union(
                self._ndarrays_cache.keys()
            )
            written_h5locs = set(
                e.h5loc for e in written_blob.values() if isinstance(e, NDArray)
            )
            missing_h5locs = ndarray_h5locs - written_h5locs
            for h5loc in missing_h5locs:
                self.log.info("Filling up %s with 0 length entry", h5loc)
                self._record_index(h5loc, 0)

        if not self.index % self.flush_frequency:
            self.flush()

        self.index += 1
        return blob

    def _record_index(self, h5loc, count, split=False):
        """Add an index entry (optionally create table) for an NDArray h5loc.

        Parameters
        ----------
        h5loc : str
            location in HDF5
        count : int
            number of elements (can be 0)
        split : bool
            if it's a split table

        """
        suffix = "/_indices" if split else "_indices"
        idx_table_h5loc = h5loc + suffix
        if idx_table_h5loc not in self.indices:
            self.indices[idx_table_h5loc] = HDF5IndexTable(
                idx_table_h5loc, start=self.index
            )

        idx_tab = self.indices[idx_table_h5loc]
        idx_tab.append(count)

    def flush(self):
        """Flush tables and arrays to disk"""
        self.log.info("Flushing tables and arrays to disk...")
        for tab in self._tables.values():
            tab.flush()
        self._write_ndarrays_cache_to_disk()

    def finish(self):
        self.flush()
        self.h5file.root._v_attrs.km3pipe = np.string_(kp.__version__)
        self.h5file.root._v_attrs.pytables = np.string_(tb.__version__)
        self.h5file.root._v_attrs.kid = np.string_(self._uuid)
        self.h5file.root._v_attrs.format_version = np.string_(FORMAT_VERSION)
        self.log.info("Adding index tables.")
        for where, idx_tab in self.indices.items():
            # any skipped NDArrays or split groups will be filled with 0 entries
            idx_tab.fillup(self.index)

            self.log.debug("Creating index table for '%s'" % where)
            h5loc = idx_tab.h5loc
            self.log.info("  -> {0}".format(h5loc))
            indices = Table(
                {"index": idx_tab.data["indices"], "n_items": idx_tab.data["n_items"]},
                h5loc=h5loc,
            )
            self._write_table(h5loc, indices, title="Indices")
        self.log.info(
            "Creating pytables index tables. " "This may take a few minutes..."
        )
        for tab in self._tables.values():
            if "frame_id" in tab.colnames:
                tab.cols.frame_id.create_index()
            if "slice_id" in tab.colnames:
                tab.cols.slice_id.create_index()
            if "dom_id" in tab.colnames:
                tab.cols.dom_id.create_index()
            if "event_id" in tab.colnames:
                try:
                    tab.cols.event_id.create_index()
                except NotImplementedError:
                    log.warning(
                        "Table '{}' has an uint64 column, "
                        "not indexing...".format(tab._v_name)
                    )
            if "group_id" in tab.colnames:
                try:
                    tab.cols.group_id.create_index()
                except NotImplementedError:
                    log.warning(
                        "Table '{}' has an uint64 column, "
                        "not indexing...".format(tab._v_name)
                    )
            tab.flush()

        if "HDF5MetaData" in self.services:
            self.log.info("Writing HDF5 meta data.")
            metadata = self.services["HDF5MetaData"]
            for name, value in metadata.items():
                self.h5file.set_node_attr("/", name, value)

        if not self.keep_open:
            self.h5file.close()
        self.cprint("HDF5 file written to: {}".format(self.filename))


class HDF5Pump(Module):
    """Read KM3NeT-formatted HDF5 files, event-by-event.

    Parameters
    ----------
    filename: str
        From where to read events. Either this OR ``filenames`` needs to be
        defined.
    skip_version_check: bool [default: False]
        Don't check the H5 version. Might lead to unintended consequences.
    shuffle: bool, optional [default: False]
        Shuffle the group_ids, so that the blobs are mixed up.
    shuffle_function: function, optional [default: np.random.shuffle
        The function to be used to shuffle the group IDs.
    reset_index: bool, optional [default: True]
        When shuffle is set to true, reset the group ID - start to count
        the group_id by 0.

    Notes
    -----
    Provides service h5singleton(h5loc): h5loc:str -> kp.Table
        Singleton tables for a given HDF5 location.
    """

    def configure(self):
        self.filename = self.get("filename")
        self.skip_version_check = self.get("skip_version_check", default=False)
        self.verbose = bool(self.get("verbose"))
        self.shuffle = self.get("shuffle", default=False)
        self.shuffle_function = self.get("shuffle_function", default=np.random.shuffle)
        self.reset_index = self.get("reset_index", default=False)

        self.h5file = None
        self.cut_mask = None
        self.indices = {}
        self._tab_indices = {}
        self._singletons = {}
        self.header = None
        self.group_ids = None
        self._n_groups = None
        self.index = 0

        self.h5file = tb.open_file(self.filename, "r")

        Provenance().record_input(self.filename, comment="HDF5Pump input")

        if not self.skip_version_check:
            check_version(self.h5file)

        self._read_group_info()

        self.expose(self.h5singleton, "h5singleton")

    def _read_group_info(self):
        h5file = self.h5file

        if "/group_info" not in h5file:
            self.log.critical("Missing /group_info '%s', aborting..." % h5file.filename)
            raise SystemExit

        self.log.info("Reading group information from '/group_info'.")
        group_info = h5file.get_node("/", "group_info")
        self.group_ids = group_info.cols.group_id[:]
        self._n_groups = len(self.group_ids)

        if "/raw_header" in h5file:
            self.log.info("Reading /raw_header")
            try:
                self.header = HDF5Header.from_pytable(h5file.get_node("/raw_header"))
            except TypeError:
                self.log.error("Could not parse the raw header, skipping!")

        if self.shuffle:
            self.log.info("Shuffling group IDs")
            self.shuffle_function(self.group_ids)

    def h5singleton(self, h5loc):
        """Returns the singleton table for a given HDF5 location"""
        return self._singletons[h5loc]

    def process(self, blob):
        self.log.info("Reading blob at index %s" % self.index)
        if self.index >= self._n_groups:
            self.log.info("All groups are read.")
            raise StopIteration
        blob = self.get_blob(self.index)
        self.index += 1
        return blob

    def get_blob(self, index):
        blob = Blob()
        group_id = self.group_ids[index]

        # skip groups with separate columns
        # and deal with them later
        # this should be solved using hdf5 attributes in near future
        split_table_locs = []
        ndarray_locs = []
        for tab in self.h5file.walk_nodes(classname="Table"):
            h5loc = tab._v_pathname
            loc, tabname = os.path.split(h5loc)
            if tabname in self.indices:
                self.log.info("index table '%s' already read, skip..." % h5loc)
                continue
            if loc in split_table_locs:
                self.log.info("get_blob: '%s' is noted, skip..." % h5loc)
                continue
            if tabname == "_indices":
                self.log.debug("get_blob: found index table '%s'" % h5loc)
                split_table_locs.append(loc)
                self.indices[loc] = self.h5file.get_node(h5loc)
                continue
            if tabname.endswith("_indices"):
                self.log.debug("get_blob: found index table '%s' for NDArray" % h5loc)
                ndarr_loc = h5loc.replace("_indices", "")
                ndarray_locs.append(ndarr_loc)
                if ndarr_loc in self.indices:
                    self.log.info(
                        "index table for NDArray '%s' already read, skip..." % ndarr_loc
                    )
                    continue
                _index_table = self.h5file.get_node(h5loc)
                self.indices[ndarr_loc] = {
                    "index": _index_table.col("index")[:],
                    "n_items": _index_table.col("n_items")[:],
                }
                continue
            tabname = camelise(tabname)

            if "group_id" in tab.dtype.names:
                try:
                    if h5loc not in self._tab_indices:
                        self._read_tab_indices(h5loc)
                    tab_idx_start = self._tab_indices[h5loc][0][group_id]
                    tab_n_items = self._tab_indices[h5loc][1][group_id]
                    if tab_n_items == 0:
                        continue
                    arr = tab[tab_idx_start : tab_idx_start + tab_n_items]
                except IndexError:
                    self.log.debug("No data for h5loc '%s'" % h5loc)
                    continue
                except NotImplementedError:
                    # 64-bit unsigned integer columns like ``group_id``
                    # are not yet supported in conditions
                    self.log.debug(
                        "get_blob: found uint64 column at '{}'...".format(h5loc)
                    )
                    arr = tab.read()
                    arr = arr[arr["group_id"] == group_id]
                except ValueError:
                    # "there are no columns taking part
                    # in condition ``group_id == 0``"
                    self.log.info(
                        "get_blob: no `%s` column found in '%s'! "
                        "skipping... " % ("group_id", h5loc)
                    )
                    continue
            else:
                if h5loc not in self._singletons:
                    log.info("Caching H5 singleton: {} ({})".format(tabname, h5loc))
                    self._singletons[h5loc] = Table(
                        tab.read(),
                        h5loc=h5loc,
                        split_h5=False,
                        name=tabname,
                        h5singleton=True,
                    )
                blob[tabname] = self._singletons[h5loc]
                continue

            self.log.debug("h5loc: '{}'".format(h5loc))
            tab = Table(arr, h5loc=h5loc, split_h5=False, name=tabname)
            if self.shuffle and self.reset_index:
                tab.group_id[:] = index
            blob[tabname] = tab

        # skipped locs are now column wise datasets (usually hits)
        # currently hardcoded, in future using hdf5 attributes
        # to get the right constructor
        for loc in split_table_locs:
            # if some events are missing (group_id not continuous),
            # this does not work as intended
            # idx, n_items = self.indices[loc][group_id]
            idx = self.indices[loc].col("index")[group_id]
            n_items = self.indices[loc].col("n_items")[group_id]
            end = idx + n_items
            node = self.h5file.get_node(loc)
            columns = (c for c in node._v_children if c != "_indices")
            data = {}
            for col in columns:
                data[col] = self.h5file.get_node(loc + "/" + col)[idx:end]
            tabname = camelise(loc.split("/")[-1])
            s_tab = Table(data, h5loc=loc, split_h5=True, name=tabname)
            if self.shuffle and self.reset_index:
                s_tab.group_id[:] = index
            blob[tabname] = s_tab

        if self.header is not None:
            blob["Header"] = self.header

        for ndarr_loc in ndarray_locs:
            self.log.info("Reading %s" % ndarr_loc)
            try:
                idx = self.indices[ndarr_loc]["index"][group_id]
                n_items = self.indices[ndarr_loc]["n_items"][group_id]
            except IndexError:
                continue
            end = idx + n_items
            ndarr = self.h5file.get_node(ndarr_loc)
            ndarr_name = camelise(ndarr_loc.split("/")[-1])
            _ndarr = NDArray(
                ndarr[idx:end], h5loc=ndarr_loc, title=ndarr.title, group_id=group_id
            )
            if self.shuffle and self.reset_index:
                _ndarr.group_id = index
            blob[ndarr_name] = _ndarr

        return blob

    def _read_tab_indices(self, h5loc):
        self.log.info("Reading table indices for '{}'".format(h5loc))
        node = self.h5file.get_node(h5loc)
        group_ids = None
        if "group_id" in node.dtype.names:
            group_ids = self.h5file.get_node(h5loc).cols.group_id[:]
        else:
            self.log.error("No data found in '{}'".format(h5loc))
            return

        self._tab_indices[h5loc] = create_index_tuple(group_ids)

    def __len__(self):
        self.log.info("Opening all HDF5 files to check the number of groups")
        n_groups = 0
        for filename in self.filenames:
            with tb.open_file(filename, "r") as h5file:
                group_info = h5file.get_node("/", "group_info")
                self.group_ids = group_info.cols.group_id[:]
                n_groups += len(self.group_ids)
        return n_groups

    def __iter__(self):
        return self

    def __next__(self):
        # TODO: wrap that in self._check_if_next_file_is_needed(self.index)
        if self.index >= self._n_groups:
            self.log.info("All groups are read")
            raise StopIteration
        blob = self.get_blob(self.index)
        self.index += 1
        return blob

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.get_blob(index)
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield self.get_blob(i)

        self.filename = None

    def _close_h5file(self):
        if self.h5file:
            self.h5file.close()

    def finish(self):
        self._close_h5file()


@jit
def create_index_tuple(group_ids):
    """An helper function to create index tuples for fast lookup in HDF5Pump"""
    max_group_id = np.max(group_ids)

    start_idx_arr = np.full(max_group_id + 1, 0)
    n_items_arr = np.full(max_group_id + 1, 0)

    current_group_id = group_ids[0]
    current_idx = 0
    item_count = 0

    for group_id in group_ids:
        if group_id != current_group_id:
            start_idx_arr[current_group_id] = current_idx
            n_items_arr[current_group_id] = item_count
            current_idx += item_count
            item_count = 0
            current_group_id = group_id
        item_count += 1
    else:
        start_idx_arr[current_group_id] = current_idx
        n_items_arr[current_group_id] = item_count

    return (start_idx_arr, n_items_arr)


class HDF5MetaData(Module):
    """Metadata to attach to the HDF5 file.

    Parameters
    ----------
    data: dict

    """

    def configure(self):
        self.data = self.require("data")
        self.expose(self.data, "HDF5MetaData")


@singledispatch
def header2table(data):
    """Convert a header to an `HDF5Header` compliant `kp.Table`"""
    print(f"Unsupported header data of type {type(data)}")


@header2table.register(dict)
def _(header_dict):
    if not header_dict:
        print("Empty header dictionary.")
        return
    tab_dict = defaultdict(list)

    for parameter, data in header_dict.items():
        fields = []
        values = []
        types = []
        for field_name, field_value in data.items():
            fields.append(field_name)
            values.append(str(field_value))
            try:
                _ = float(field_value)  # noqa
                types.append("f4")
            except ValueError:
                types.append("a{}".format(len(field_value)))
            except TypeError:  # e.g. values is None
                types.append("a{}".format(len(str(field_value))))
        tab_dict["parameter"].append(parameter.encode())
        tab_dict["field_names"].append(" ".join(fields).encode())
        tab_dict["field_values"].append(" ".join(values).encode())
        tab_dict["dtype"].append(" ".join(types).encode())
        log.debug(
            "{}: {} {} {}".format(
                tab_dict["parameter"][-1],
                tab_dict["field_names"][-1],
                tab_dict["field_values"][-1],
                tab_dict["dtype"][-1],
            )
        )
    return Table(tab_dict, h5loc="/raw_header", name="RawHeader", h5singleton=True)


@header2table.register(km3io.offline.Header)
def _(header):
    out = {}
    for parameter, values in header._data.items():
        try:
            values = values._asdict()
        except AttributeError:
            # single entry without further parameter name
            # in specification
            values = {parameter + "_0": values}
        out[parameter] = values
    return header2table(out)


@header2table.register(HDF5Header)
def _(header):
    return header2table(header._data)



"""
A collection of utils.

"""

from km3pipe.utils import h5info

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Filename: qprefit.py
# Author: Tamas Gal <tgal@km3net.de>
"""
======================
ROyPrefit batch script
======================

Apply the ROyPrefit to a set of files.

Usage:
    qprefit.py PATH OUTPUT_PATH [options] FITTER
    qprefit.py (-h | --help)

Options:
    PATH         Path to the files.
    FITTER       The fitter script, which will be called with the filepath.
    OUTPUT_PATH  Path to store the prefit summary CSV files.
    -x SUFFIX    Suffix to append to the output file [default: .royprefit.csv].
    -n N_FILES   Number of files to process per job [default: 20].
    -e ET        Estimated walltime per file in minutes [default: 3].
    -m VMEM      Estimated vmem for a job [default: 8G].
    -j JOBNAME   The name of the submitted jobs [default: royprefit].
    -l LOG_PATH  Path of the job log files [default: qlogs].
    -q           Dryrun: don't submit jobs, just print the first job script.
    -h --help    Show this screen.

"""

from glob import glob
import os
from os.path import basename, join, abspath
import time
import km3pipe as kp
from km3pipe.tools import chunks

__author__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__version__ = "1.0"


def mkdir(path):
    """Create folder if it does not exist."""
    if not os.path.exists(path):
        os.makedirs(path)


def main():
    from docopt import docopt

    args = docopt(__doc__, version=__version__)

    CWD = os.getcwd()
    FITTER = abspath(args["FITTER"])
    PATH = abspath(args["PATH"])
    OUTPUT_PATH = join(CWD, args["OUTPUT_PATH"])
    SUFFIX = args["-x"]
    ET_PER_FILE = int(args["-e"]) * 60  # [s]
    FILES_PER_JOB = int(args["-n"])
    VMEM = args["-m"]
    LOG_PATH = args["-l"]
    JOB_NAME = args["-j"]
    DRYRUN = args["-q"]

    mkdir(OUTPUT_PATH)

    files = glob(join(PATH, "*.h5"))
    summaries = [
        basename(f)[: -len(SUFFIX)] for f in glob(join(OUTPUT_PATH, "*" + SUFFIX))
    ]
    rem_files = list(set(basename(f) for f in files) - set(summaries))

    print("{} files in total".format(len(files)))
    print("{} files to be fitted.".format(len(rem_files)))

    s = kp.shell.Script()

    for job_id, file_chunk in enumerate(chunks(rem_files, FILES_PER_JOB)):
        n_files = len(file_chunk)
        s.add("echo 'ROyPrefitting {} files'".format(n_files))
        s.add("cd $TMPDIR; mkdir -p $USER; cd $USER")
        s.add("echo")

        for fname in file_chunk:
            fpath = join(PATH, fname)
            oname = fname + SUFFIX
            opath = join(OUTPUT_PATH, oname)
            s.add("echo '" + 42 * "=" + "'")
            s.add("echo Processing {}:".format(fname))
            s.add("{} {}".format(FITTER, fpath))
            s.add("cp {} {}".format(oname, opath))
            s.add("echo File '{}' fitted.".format(fname))
            s.add("echo '" + 42 * "=" + "'")

        walltime = time.strftime("%H:%M:%S", time.gmtime(ET_PER_FILE * n_files))

        kp.shell.qsub(
            s,
            "{}_{}".format(JOB_NAME, job_id),
            walltime=walltime,
            vmem=VMEM,
            log_path=LOG_PATH,
            irods=True,
            dryrun=DRYRUN,
        )

        if DRYRUN:
            break

        s.clear()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# vim: ts=4 sw=4 et

from itertools import combinations
import os

import numpy as np

import km3pipe as kp
from km3modules.plot import plot_dom_parameters
from km3modules.fit import fit_delta_ts
import km3pipe.style

km3pipe.style.use('km3pipe')

PLOTS_PATH = 'www/plots'
cal = kp.calib.Calibration(det_id=14)
detector = cal.detector


def mongincidence(times, tdcs, tmax=20, offset=0):
    coincidences = []
    cur_t = 0
    las_t = 0
    for t_idx, t in enumerate(times):
        cur_t = t
        diff = cur_t - las_t
        if offset < diff <= offset + tmax \
                and t_idx > 0 \
                and tdcs[t_idx - 1] != tdcs[t_idx]:
            coincidences.append(((tdcs[t_idx - 1], tdcs[t_idx]), diff))
        las_t = cur_t
    return coincidences


class MonitorK40(kp.Module):
    def configure(self):
        self.index = 0
        self.k40_2fold = {}
        self.rates = {}
        self.cuckoo = kp.time.Cuckoo(300, self.create_plot)
        self.n_slices = 0

    def process(self, blob):
        self.index += 1
        #        if self.index % 30:
        #            return blob

        self.n_slices += 1

        for dom_id, hits in blob['TimesliceFrames'].iteritems():
            du, dom, _ = detector.doms[dom_id]
            omkey = (du, dom)
            if omkey not in self.rates:
                self.rates[omkey] = np.zeros(shape=(465, 41))
            hits.sort(key=lambda x: x[1])
            times = [t for (_, t, _) in hits]
            tdcs = [t for (t, _, _) in hits]
            coinces = mongincidence(times, tdcs)
            combs = list(combinations(range(31), 2))
            for pmt_pair, t in coinces:
                if pmt_pair[0] > pmt_pair[1]:
                    pmt_pair = (pmt_pair[1], pmt_pair[0])
                    t = -t
                self.rates[omkey][combs.index(pmt_pair), t + 20] += 1

        self.cuckoo.msg()

        return blob

    def create_plot(self):
        print(self.__class__.__name__ + ": updating plot.")

        data = {}
        for omkey, m in self.rates.iteritems():
            print("Fitting {0}".format(omkey))
            time = self.n_slices / 10
            rates, means = fit_delta_ts(m, time)
            overall_rate = np.sum(rates)
            print(overall_rate)
            data[omkey] = overall_rate
        self.rates = {}
        self.n_slices = 0

        filename = os.path.join(PLOTS_PATH, 'k40.png')
        plot_dom_parameters(
            data,
            detector,
            filename,
            'rate [Hz]',
            "K40 rate",
            vmin=600,
            vmax=1200,
            cmap='coolwarm',
            missing='black',
            under='darkorchid',
            over='deeppink'
        )
        print("done")


pipe = kp.Pipeline()
pipe.attach(
    kp.io.ch.CHPump,
    host='127.0.0.1',
    port=5553,
    tags='IO_TSL',
    timeout=60 * 60 * 24 * 7,
    max_queue=2000
)
pipe.attach(kp.io.daq.TimesliceParser)
pipe.attach(MonitorK40)
pipe.drain()



#!/usr/bin/env python3

#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A test runner for Jpp.
This script traverses the tests directory and reports a summary.

"""
from __future__ import print_function
from collections import defaultdict
from glob import glob, iglob
import os
from os.path import join, basename
from subprocess import Popen, PIPE
import sys
from time import time
import re
import xml.etree.ElementTree as ET
import xml.dom.minidom

if not len(sys.argv) == 2:
    print("Usage: run_tests.py PATH_TO_TESTS")
    raise SystemExit

os.environ["TEST_DEBUG"] = "1"

PY2 = sys.version_info < (3, 0, 0)

try:
    unichr
except NameError:
    unichr = chr

__author__ = "Tamas Gal"
__credits__ = "Brian Beyer"
__license__ = "MIT"
__email__ = "tgal@km3net.de"
__status__ = "Development"

TESTS_DIR = sys.argv[1]
JUNIT_XML = 'reports/junit_{}.xml'.format(os.path.basename(TESTS_DIR))

if hasattr(sys.stdout, 'isatty') and sys.stdout.isatty():
    INFO = '\033[94m'  # blue
    OK = '\033[92m'  # green
    FAIL = '\033[91m'  # red
    RST = '\033[0m'  # colour reset
    BOLD = '\033[1m'
else:
    INFO, OK, FAIL, RST, BOLD = ('', ) * 5


def main():
    test_results = run_tests(TESTS_DIR)
    n_tests = len(test_results)
    n_failed_tests = sum(1 for r in test_results.values() if r[0] > 0)
    total_time = sum(r[1] for r in test_results.values())

    print("\n{}"
          "Test summary\n"
          "============\n{}"
          "Total number of tests: {}\n{}"
          "Failed tests: {}{}\n"
          "Elapsed time: {:.1f}s\n".format(
              INFO, RST, n_tests, BOLD + FAIL if n_failed_tests > 0 else OK,
              n_failed_tests, RST, total_time))

    write_junit_xml(test_results)

    if n_failed_tests > 0:
        print_captured_output(test_results)
        exit(1)
    else:
        exit(0)


def write_junit_xml(test_results):
    """Generate XML file according to JUnit specs"""
    test_cases = []
    for test_script, (exit_code, t, stdout, stderr) in test_results.items():
        test_case = TestCase(test_script,
                             elapsed_sec=t,
                             stdout=stdout,
                             stderr=stderr)
        if exit_code > 0:
            test_case.add_error_info('non-zero exit-code: %d' % exit_code)
        test_cases.append(test_case)
    test_suite = TestSuite("Jpp Test Suite", test_cases)
    with open(JUNIT_XML, 'w') as f:
        TestSuite.to_file(f, [test_suite])


def print_captured_output(test_results):
    """Prints the STDOUT and STDERR of failing test scripts"""
    print("{}"
          "Captured output of failing tests\n"
          "================================\n{}".format(INFO, RST))
    for test_script, (exit_code, t, stdout, stderr) in test_results.items():
        if exit_code > 0:
            print("{}\n{}\n".format(test_script, len(test_script) * '-'))
            print('{}stdout:{}\n{}\n{}stderr:{}\n{}'.format(
                OK + BOLD, RST, stdout.decode('utf-8'), FAIL + BOLD, RST,
                stderr.decode('utf-8')))


def run_tests(tests_dir):
    """Runs each script in the tests directory and returns the results.

    Parameters
    ----------
    tests_dir: str
      The path to the test dir, containing the test scripts (`*.sh`).

    Returns
    -------
    dict: key = script path, value = (exit_code, elapsed_time, stdout, stderr)

    """
    test_results = {}

    for subdir in sorted(glob(join(tests_dir, '*'))):
        component_group = basename(subdir)
        print("\n{}{}\n{}{}".format(INFO, component_group,
                                    len(component_group) * '=', RST))
        for test_script in sorted(glob(join(subdir, '*.*'))):
            print("+ {}".format(test_script), end=' => ')
            sys.stdout.flush()
            start_time = time()
            proc = Popen(test_script, stdout=PIPE, stderr=PIPE)
            out, err = proc.communicate()
            exit_code = proc.wait()
            delta_t = time() - start_time
            test_results[test_script] = (exit_code, delta_t, out, err)
            print(" ({:.2f} s) ".format(delta_t), end='')
            sys.stdout.flush()
            if exit_code > 0:
                print("{}FAILED (exit code {}){}".format(FAIL, exit_code, RST))
                sys.stdout.flush()
            else:
                print("{}OK{}".format(OK, RST))
                sys.stdout.flush()

    return test_results


def decode(var, encoding):
    """
    If not already unicode, decode it.
    """
    if PY2:
        if isinstance(var, unicode):
            ret = var
        elif isinstance(var, str):
            if encoding:
                ret = var.decode(encoding)
            else:
                ret = unicode(var)
        else:
            ret = unicode(var)
    else:
        ret = str(var)
    return ret


class TestSuite(object):
    """
    Suite of test cases.
    Can handle unicode strings or binary strings if their encoding is provided.
    """
    def __init__(self,
                 name,
                 test_cases=None,
                 hostname=None,
                 id=None,
                 package=None,
                 timestamp=None,
                 properties=None,
                 file=None,
                 log=None,
                 url=None,
                 stdout=None,
                 stderr=None):
        self.name = name
        if not test_cases:
            test_cases = []
        try:
            iter(test_cases)
        except TypeError:
            raise Exception('test_cases must be a list of test cases')
        self.test_cases = test_cases
        self.timestamp = timestamp
        self.hostname = hostname
        self.id = id
        self.package = package
        self.file = file
        self.log = log
        self.url = url
        self.stdout = stdout
        self.stderr = stderr
        self.properties = properties

    def build_xml_doc(self, encoding=None):
        """
        Builds the XML document for the JUnit test suite.
        Produces clean unicode strings and decodes non-unicode with the help of encoding.
        @param encoding: Used to decode encoded strings.
        @return: XML document with unicode string elements
        """

        # build the test suite element
        test_suite_attributes = dict()
        test_suite_attributes['name'] = decode(self.name, encoding)
        if any(c.assertions for c in self.test_cases):
            test_suite_attributes['assertions'] = \
                str(sum([int(c.assertions) for c in self.test_cases if c.assertions]))
        test_suite_attributes['disabled'] = \
            str(len([c for c in self.test_cases if not c.is_enabled]))
        test_suite_attributes['failures'] = \
            str(len([c for c in self.test_cases if c.is_failure()]))
        test_suite_attributes['errors'] = \
            str(len([c for c in self.test_cases if c.is_error()]))
        test_suite_attributes['skipped'] = \
            str(len([c for c in self.test_cases if c.is_skipped()]))
        test_suite_attributes['time'] = \
            str(sum(c.elapsed_sec for c in self.test_cases if c.elapsed_sec))
        test_suite_attributes['tests'] = str(len(self.test_cases))

        if self.hostname:
            test_suite_attributes['hostname'] = decode(self.hostname, encoding)
        if self.id:
            test_suite_attributes['id'] = decode(self.id, encoding)
        if self.package:
            test_suite_attributes['package'] = decode(self.package, encoding)
        if self.timestamp:
            test_suite_attributes['timestamp'] = decode(
                self.timestamp, encoding)
        if self.file:
            test_suite_attributes['file'] = decode(self.file, encoding)
        if self.log:
            test_suite_attributes['log'] = decode(self.log, encoding)
        if self.url:
            test_suite_attributes['url'] = decode(self.url, encoding)

        xml_element = ET.Element("testsuite", test_suite_attributes)

        # add any properties
        if self.properties:
            props_element = ET.SubElement(xml_element, "properties")
            for k, v in self.properties.items():
                attrs = {
                    'name': decode(k, encoding),
                    'value': decode(v, encoding)
                }
                ET.SubElement(props_element, "property", attrs)

        # add test suite stdout
        if self.stdout:
            stdout_element = ET.SubElement(xml_element, "system-out")
            stdout_element.text = decode(self.stdout, encoding)

        # add test suite stderr
        if self.stderr:
            stderr_element = ET.SubElement(xml_element, "system-err")
            stderr_element.text = decode(self.stderr, encoding)

        # test cases
        for case in self.test_cases:
            test_case_attributes = dict()
            test_case_attributes['name'] = decode(case.name, encoding)
            if case.assertions:
                # Number of assertions in the test case
                test_case_attributes['assertions'] = "%d" % case.assertions
            if case.elapsed_sec:
                test_case_attributes['time'] = "%f" % case.elapsed_sec
            if case.timestamp:
                test_case_attributes['timestamp'] = decode(
                    case.timestamp, encoding)
            if case.classname:
                test_case_attributes['classname'] = decode(
                    case.classname, encoding)
            if case.status:
                test_case_attributes['status'] = decode(case.status, encoding)
            if case.category:
                test_case_attributes['class'] = decode(case.category, encoding)
            if case.file:
                test_case_attributes['file'] = decode(case.file, encoding)
            if case.line:
                test_case_attributes['line'] = decode(case.line, encoding)
            if case.log:
                test_case_attributes['log'] = decode(case.log, encoding)
            if case.url:
                test_case_attributes['url'] = decode(case.url, encoding)

            test_case_element = ET.SubElement(xml_element, "testcase",
                                              test_case_attributes)

            # failures
            if case.is_failure():
                attrs = {'type': 'failure'}
                if case.failure_message:
                    attrs['message'] = decode(case.failure_message, encoding)
                if case.failure_type:
                    attrs['type'] = decode(case.failure_type, encoding)
                failure_element = ET.Element("failure", attrs)
                if case.failure_output:
                    failure_element.text = decode(case.failure_output,
                                                  encoding)
                test_case_element.append(failure_element)

            # errors
            if case.is_error():
                attrs = {'type': 'error'}
                if case.error_message:
                    attrs['message'] = decode(case.error_message, encoding)
                if case.error_type:
                    attrs['type'] = decode(case.error_type, encoding)
                error_element = ET.Element("error", attrs)
                if case.error_output:
                    error_element.text = decode(case.error_output, encoding)
                test_case_element.append(error_element)

            # skippeds
            if case.is_skipped():
                attrs = {'type': 'skipped'}
                if case.skipped_message:
                    attrs['message'] = decode(case.skipped_message, encoding)
                skipped_element = ET.Element("skipped", attrs)
                if case.skipped_output:
                    skipped_element.text = decode(case.skipped_output,
                                                  encoding)
                test_case_element.append(skipped_element)

            # test stdout
            if case.stdout:
                stdout_element = ET.Element("system-out")
                stdout_element.text = decode(case.stdout, encoding)
                test_case_element.append(stdout_element)

            # test stderr
            if case.stderr:
                stderr_element = ET.Element("system-err")
                stderr_element.text = decode(case.stderr, encoding)
                test_case_element.append(stderr_element)

        return xml_element

    @staticmethod
    def to_xml_string(test_suites, prettyprint=True, encoding=None):
        """
        Returns the string representation of the JUnit XML document.
        @param encoding: The encoding of the input.
        @return: unicode string
        """

        try:
            iter(test_suites)
        except TypeError:
            raise Exception('test_suites must be a list of test suites')

        xml_element = ET.Element("testsuites")
        attributes = defaultdict(int)
        for ts in test_suites:
            ts_xml = ts.build_xml_doc(encoding=encoding)
            for key in ['failures', 'errors', 'tests', 'disabled']:
                attributes[key] += int(ts_xml.get(key, 0))
            for key in ['time']:
                attributes[key] += float(ts_xml.get(key, 0))
            xml_element.append(ts_xml)
        for key, value in attributes.items():
            xml_element.set(key, str(value))

        xml_string = ET.tostring(xml_element, encoding=encoding)
        # is encoded now
        xml_string = TestSuite._clean_illegal_xml_chars(
            xml_string.decode(encoding or 'utf-8'))
        # is unicode now

        if prettyprint:
            # minidom.parseString() works just on correctly encoded binary strings
            xml_string = xml_string.encode(encoding or 'utf-8')
            xml_string = xml.dom.minidom.parseString(xml_string)
            # toprettyxml() produces unicode if no encoding is being passed or binary string with an encoding
            xml_string = xml_string.toprettyxml(encoding=encoding)
            if encoding:
                xml_string = xml_string.decode(encoding)
            # is unicode now
        return xml_string

    @staticmethod
    def to_file(file_descriptor, test_suites, prettyprint=True, encoding=None):
        """
        Writes the JUnit XML document to a file.
        """
        xml_string = TestSuite.to_xml_string(test_suites,
                                             prettyprint=prettyprint,
                                             encoding=encoding)
        # has problems with encoded str with non-ASCII (non-default-encoding) characters!
        file_descriptor.write(xml_string)

    @staticmethod
    def _clean_illegal_xml_chars(string_to_clean):
        """
        Removes any illegal unicode characters from the given XML string.
        @see: http://stackoverflow.com/questions/1707890/fast-way-to-filter-illegal-xml-unicode-chars-in-python
        """

        illegal_unichrs = [(0x00, 0x08), (0x0B, 0x1F), (0x7F, 0x84),
                           (0x86, 0x9F), (0xD800, 0xDFFF), (0xFDD0, 0xFDDF),
                           (0xFFFE, 0xFFFF), (0x1FFFE, 0x1FFFF),
                           (0x2FFFE, 0x2FFFF), (0x3FFFE, 0x3FFFF),
                           (0x4FFFE, 0x4FFFF), (0x5FFFE, 0x5FFFF),
                           (0x6FFFE, 0x6FFFF), (0x7FFFE, 0x7FFFF),
                           (0x8FFFE, 0x8FFFF), (0x9FFFE, 0x9FFFF),
                           (0xAFFFE, 0xAFFFF), (0xBFFFE, 0xBFFFF),
                           (0xCFFFE, 0xCFFFF), (0xDFFFE, 0xDFFFF),
                           (0xEFFFE, 0xEFFFF), (0xFFFFE, 0xFFFFF),
                           (0x10FFFE, 0x10FFFF)]

        illegal_ranges = [
            "%s-%s" % (unichr(low), unichr(high))
            for (low, high) in illegal_unichrs if low < sys.maxunicode
        ]

        illegal_xml_re = re.compile('[%s]' % ''.join(illegal_ranges))
        return illegal_xml_re.sub('', string_to_clean)


class TestCase(object):
    """A JUnit test case with a result and possibly some stdout or stderr"""
    def __init__(self,
                 name,
                 classname=None,
                 elapsed_sec=None,
                 stdout=None,
                 stderr=None,
                 assertions=None,
                 timestamp=None,
                 status=None,
                 category=None,
                 file=None,
                 line=None,
                 log=None,
                 group=None,
                 url=None):
        self.name = name
        self.assertions = assertions
        self.elapsed_sec = elapsed_sec
        self.timestamp = timestamp
        self.classname = classname
        self.status = status
        self.category = category
        self.file = file
        self.line = line
        self.log = log
        self.url = url
        self.stdout = stdout
        self.stderr = stderr

        self.is_enabled = True
        self.error_message = None
        self.error_output = None
        self.error_type = None
        self.failure_message = None
        self.failure_output = None
        self.failure_type = None
        self.skipped_message = None
        self.skipped_output = None

    def add_error_info(self, message=None, output=None, error_type=None):
        """Adds an error message, output, or both to the test case"""
        if message:
            self.error_message = message
        if output:
            self.error_output = output
        if error_type:
            self.error_type = error_type

    def add_failure_info(self, message=None, output=None, failure_type=None):
        """Adds a failure message, output, or both to the test case"""
        if message:
            self.failure_message = message
        if output:
            self.failure_output = output
        if failure_type:
            self.failure_type = failure_type

    def add_skipped_info(self, message=None, output=None):
        """Adds a skipped message, output, or both to the test case"""
        if message:
            self.skipped_message = message
        if output:
            self.skipped_output = output

    def is_failure(self):
        """returns true if this test case is a failure"""
        return self.failure_output or self.failure_message

    def is_error(self):
        """returns true if this test case is an error"""
        return self.error_output or self.error_message

    def is_skipped(self):
        """returns true if this test case has been skipped"""
        return self.skipped_output or self.skipped_message


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Filename: online.py
# pylint: disable=
"""
Pump for the online file format 
(the file format formerly known as formerly as JPP)

"""

import numpy as np
import km3io

from thepipe import Module, Blob
from km3pipe.dataclasses import Table
from km3pipe.logger import get_logger

log = get_logger(__name__)  # pylint: disable=C0103

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = ["Thomas Heid", "Giuliano Maggi", "Moritz Lotze", "Johannes Schumann"]
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class EventPump(Module):
    """A pump for DAQ (triggered) events in online files.

    Parameters
    ----------
    filename: str
        Name of the file to open.

    """

    def configure(self):
        self.event_index = self.get("index") or 0
        self.filename = self.require("filename")

        self.event_reader = km3io.OnlineReader(self.filename.encode())
        self.blobs = self.blob_generator()
        self.n_events = len(self.event_reader.events)
        self._current_blob = Blob()

    def blob_generator(self):
        for i in range(self.n_events):
            try:
                yield self.extract_event(i)
            except IndexError:
                return

    def __getitem__(self, index):
        if index >= self.n_events:
            raise IndexError
        return self.extract_event(index)

    def _get_trigger_mask(self, snapshot_hits, triggered_hits):
        trg_mask = np.zeros(len(snapshot_hits))
        s = np.array(
            [snapshot_hits.time, snapshot_hits.channel_id, snapshot_hits.dom_id]
        ).transpose()
        t = np.array(
            [triggered_hits.time, triggered_hits.channel_id, triggered_hits.dom_id]
        ).transpose()
        cmp_mask = s == t[:, None]
        trg_map = np.argwhere(np.all(cmp_mask, axis=2))[:, 1]
        trg_mask[trg_map] = triggered_hits.trigger_mask
        return trg_mask

    def extract_event(self, event_number):
        blob = self._current_blob
        r = self.event_reader
        hits = r.events.snapshot_hits[event_number]
        trg_hits = r.events.triggered_hits[event_number]
        raw_event_info = r.events.headers[event_number]

        trigger_mask = self._get_trigger_mask(hits, trg_hits)
        hit_series = Table(
            {
                "channel_id": hits.channel_id,
                "dom_id": hits.dom_id,
                "time": hits.time,
                "tot": hits.tot,
                "triggered": trigger_mask,
            },
            name="Hits",
            h5loc="/hits",
            split_h5=True,
        )

        event_info = Table(
            {
                "det_id": raw_event_info["detector_id"],
                "frame_index": raw_event_info["frame_index"],
                "livetime_sec": 0,
                "mc_id": 0,
                "mc_t": 0,
                "n_events_gen": 0,
                "n_files_gen": 0,
                "overlays": raw_event_info["overlays"],
                "trigger_counter": raw_event_info["trigger_counter"],
                "trigger_mask": raw_event_info["trigger_mask"],
                "utc_nanoseconds": raw_event_info["UTC_16nanosecondcycles"] * 16.0,
                "utc_seconds": raw_event_info["UTC_seconds"],
                "weight_w1": np.nan,
                "weight_w2": np.nan,
                "weight_w3": np.nan,
                "run_id": raw_event_info["run"],
            },
            name="EventInfo",
            h5loc="/event_info",
        )

        self.event_index += 1
        blob["EventInfo"] = event_info
        blob["Hits"] = hit_series
        return blob

    def process(self, blob):
        self._current_blob = blob
        return next(self.blobs)

    def __iter__(self):
        return self

    def __next__(self):
        self._current_blob = next(self.blobs)
        return self._current_blob


class TimeslicePump(Module):
    """A pump to read and extract timeslices. Currently only hits are read.

    Parameters
    -------------------
    filename: str
    stream: str ('L0', 'L1', 'L2', 'SN') default: 'L1'

    """

    def configure(self):
        fname = self.require("filename")
        self.stream = self.get("stream", default="L1")
        self.blobs = self.timeslice_generator()
        self.r = km3io.OnlineReader(fname)
        self.timeslice_info = self.create_timeslice_info()
        self.n_timeslices = len(self.timeslice_info)

        self._current_blob = Blob()
        self._hits_blob_key = "{}Hits".format(self.stream if self.stream else "TS")

    def create_timeslice_info(self):
        header = self.r.timeslices.stream(self.stream, 0).header
        slice_ids = header["frame_index"].array()
        timestamps = header["timeslice_start.UTC_seconds"].array()
        number_of_slices = len(slice_ids)
        nanoseconds = header["timeslice_start.UTC_16nanosecondcycles"].array() * 16
        timeslice_info = Table.from_template(
            {
                "frame_index": slice_ids,
                "slice_id": range(number_of_slices),
                "timestamp": timestamps,
                "nanoseconds": nanoseconds,
                "n_frames": np.zeros(len(slice_ids)),
            },
            "TimesliceInfo",
        )
        return timeslice_info

    def process(self, blob):
        self._current_blob = blob
        return next(self.blobs)

    def timeslice_generator(self):
        """Uses slice ID as iterator"""
        slice_id = 0
        while slice_id < self.n_timeslices:
            blob = self.get_blob(slice_id)
            yield blob
            slice_id += 1

    def get_blob(self, index):
        """Index is slice ID"""
        blob = self._current_blob
        hits = self._extract_hits(index)
        hits.group_id = index
        blob["TimesliceInfo"] = self.timeslice_info[index : index + 1]
        blob["TimesliceInfo"]["n_frames"] = self._extract_number_of_frames(index)
        blob[self._hits_blob_key] = hits
        return blob

    def _extract_number_of_frames(self, index):
        timeslice = self.r.timeslices.stream(self.stream, index)
        return len(timeslice.frames)

    def _extract_hits(self, index):
        timeslice = self.r.timeslices.stream(self.stream, index)
        raw_hits = {
            "dom_id": [],
            "channel_id": [],
            "time": [],
            "tot": [],
            "group_id": [],
        }

        for dom_id, frame in timeslice.frames.items():
            raw_hits["channel_id"].extend(frame.pmt)
            raw_hits["time"].extend(frame.tdc)
            raw_hits["tot"].extend(frame.tot)
            raw_hits["dom_id"].extend(len(frame.pmt) * [dom_id])
            raw_hits["group_id"].extend(len(frame.pmt) * [0])

        hits = Table.from_template(raw_hits, "TimesliceHits")
        return hits

    def get_by_frame_index(self, frame_index):
        blob = Blob()
        ts_info = self.timeslice_info[self.timeslice_info.frame_index == frame_index][0]
        slice_id = ts_info.slice_id
        hits = self._extract_hits(slice_id)
        blob[self._hits_blob_key] = hits
        return blob

    def __len__(self):
        return self.n_timeslices

    def __iter__(self):
        return self

    def __next__(self):
        self._current_blob = next(self.blobs)
        return self._current_blob

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.get_blob(index)
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield self.get_blob(i)



# Filename: h5header.py
"""
Show the km3pipe etc. version used to write a H5 file.

Usage:
    h5header FILE [-r]
    h5header (-h | --help)
    h5header --version

Options:
    FILE        Input file.
    -r --raw    Dump raw metadata.
    -h --help   Show this screen.

"""

import tables

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__credits__ = "Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def h5header(h5name, raw=False):
    with tables.open_file(h5name) as h5:
        header_loc = "/header"
        if not header_loc in h5:  # noqa
            print("Sorry, no header.")
            return
        header = h5.get_node(header_loc)
        att_list = header._v_attrs._f_list()
        if raw:
            print(header._v_attrs.__repr__())
            return
        if not att_list and not raw:
            print("Sorry, no metadata.")
            return
        for att in header._v_attrs._f_list():
            ver = header._v_attrs[att]
            print("{}: {}".format(att, ver))


def main():
    from docopt import docopt

    arguments = docopt(__doc__)

    h5header(arguments["FILE"], arguments["--raw"])



#!/usr/bin/env python3
# Filename: daqsample.py
"""
Take samples from a given DAQ ControlHost stream.

Usage:
    daqsample [options] TAG OUTFILE
    daqsample (-h | --help)
    daqsample --version

Options:
    TAG             ControlHost TAG (e.g. IO_EVT).
    OUTFILE         Filename of the dump file.
    -i LIGIER_IP    IP of the Ligier [default: 127.0.0.1].
    -p LIGIER_PORT  Port of the Ligier [default: 5553].
    -n SAMPLES      Number of samples [default: 1].
    -h --help       Show this screen.

"""
import km3pipe as kp

__author__ = "Tamas Gal"
__copyright__ = "Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class Dumper(kp.Module):
    def configure(self):
        filename = self.require("filename")
        self._fobj = open(filename, "bw")

    def process(self, blob):
        self._fobj.write(blob["CHData"])
        return blob

    def finish(self):
        self._fobj.close()


def main():
    from docopt import docopt

    args = docopt(__doc__, version=kp.version)

    tag = args["TAG"]
    outfile = args["OUTFILE"]
    port = int(args["-p"])
    ip = args["-i"]
    n = int(args["-n"])

    pipe = kp.Pipeline()
    pipe.attach(kp.io.ch.CHPump, host=ip, port=port, tags=tag)
    pipe.attach(Dumper, filename=outfile)
    pipe.drain(n)



# Filename: __init__.py
"""
A collection of io for different kinds of data formats.

"""

import os.path

import numpy as np

from .evt import EvtPump  # noqa
from .clb import CLBPump  # noqa
from .ch import CHPump  # noqa
from .hdf5 import HDF5Pump, HDF5Sink, HDF5MetaData  # noqa
from .offline import OfflinePump
from . import online
from . import offline
from . import daq

from km3pipe.logger import get_logger

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal, Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"



# Filename: test_shell.py
# pylint: disable=locally-disabled,C0111,R0904,C0103

from km3pipe.testing import TestCase
from km3pipe.shell import Script

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2017, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestScript(TestCase):
    def test_add(self):
        s = Script()
        s.add("a")
        s.add("b")

    def test_str(self):
        s = Script()
        s.add("a")
        s.add("b")
        self.assertEqual("a\nb", str(s))

    def test_clear(self):
        s = Script()
        s.add("a")
        s.clear()
        self.assertEqual("", str(s))

    def test_add_two_argument_command(self):
        s = Script()
        s._add_two_argument_command("command", "a", "b")
        self.assertEqual("command a b", str(s))

    def test_add_two_argument_command_multiple_times(self):
        s = Script()
        s._add_two_argument_command("cmd1", "a", "b")
        s._add_two_argument_command("cmd2", "c", "d")
        s._add_two_argument_command("cmd3", "e", "f")
        self.assertEqual("cmd1 a b\ncmd2 c d\ncmd3 e f", str(s))

    def test_cp(self):
        s = Script()
        s.cp("a", "b")
        self.assertEqual("cp a b", str(s))

    def test_mv(self):
        s = Script()
        s.mv("a", "b")
        self.assertEqual("mv a b", str(s))

    def test_echo(self):
        s = Script()
        s.echo("test")
        self.assertEqual('echo "test"', str(s))

    def test_separator(self):
        s = Script()
        s.separator()
        self.assertEqual('echo "' + "=" * 42 + '"', str(s))

    def test_mkdir(self):
        s = Script()
        s.mkdir("/path/to/file")
        self.assertEqual('mkdir -p "/path/to/file"', str(s))

    def test_iget(self):
        s = Script()
        s.iget("/path/to/file")
        self.assertEqual('iget -v "/path/to/file"', str(s))

    def test_combining_scripts(self):
        s1 = Script()
        s2 = Script()
        s1.add("a")
        s1.add("b")
        s2.add("c")
        s2.add("d")
        assert "a\nb\nc\nd" == str(s1 + s2)



#!/usr/bin/env python
"""
Convert jupyter notebook to sphinx gallery notebook styled examples.
Needs pypandoc (`pip install pypandoc`).

Usage:
    nb2sphx NOTEBOOK

"""

import json
import km3pipe.extras

from km3pipe import version


def convert_ipynb_to_gallery(file_name):
    """
    Blatantly stolen + adapted from
    https://gist.github.com/wuhuikai/4a7ceb8bc52454e17a4eb8327d538d85

    """
    pdoc = km3pipe.extras.pypandoc()

    python_file = ""

    nb_dict = json.load(open(file_name))
    cells = nb_dict["cells"]

    for i, cell in enumerate(cells):
        if i == 0:
            assert cell["cell_type"] == "markdown", "First cell has to be markdown"

            md_source = "".join(cell["source"])
            rst_source = pdoc.convert_text(md_source, "rst", "md")
            python_file = '"""\n' + rst_source + '\n"""'
        else:
            if cell["cell_type"] == "markdown":
                md_source = "".join(cell["source"])
                rst_source = pdoc.convert_text(md_source, "rst", "md")
                commented_source = "\n".join(["# " + x for x in rst_source.split("\n")])
                python_file = (
                    python_file + "\n\n\n" + "#" * 70 + "\n" + commented_source
                )
            elif cell["cell_type"] == "code":
                source = "".join(cell["source"])
                python_file = python_file + "\n" * 2 + source

    open(file_name.replace(".ipynb", ".py"), "w").write(python_file)


def main():
    from docopt import docopt

    args = docopt(__doc__, version=version)
    fname = args["NOTEBOOK"]
    convert_ipynb_to_gallery(fname)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
========
HDF5Sink
========

This examples show how to create dummy (event) data, dump them to
an HDF5 file using the ``kp.io.HDF5Sink`` module and how to read them back
with pandas or the ``kp.io.HDF5Pump`` module.

"""

import km3pipe as kp
import pandas as pd
import numpy as np

filename = "reco_results.h5"

####################################################################
# The ``DataGenerator`` is a dumm pump which creates random event
# data and a common header.


class DataGenerator(kp.Module):
    def configure(self):
        self.header = kp.Table({"is_mc": True}, h5singleton=True, h5loc="/header")

    def process(self, blob):

        n = np.random.randint(2, 20)
        tab = kp.Table(
            {"lambda": np.random.rand(n), "energy": np.random.rand(n) * 1000},
            h5loc="/reco",
            name="Quality Parameter",
        )
        event_info = kp.Table(
            {"run_id": np.random.randint(10000, 20000)}, h5loc="/event_info"
        )
        blob["Reco"] = tab
        blob["EventInfo"] = event_info
        blob["Header"] = self.header
        return blob


####################################################################
# The ``printer`` just prints the blob, obviously.
# data and a common header.


def printer(blob):
    print(blob)
    return blob


####################################################################
# The following pipeline will do 10 iterations and dump the data
# properly formatted into our output file, ``reso_results.h5``

pipe = kp.Pipeline()
pipe.attach(DataGenerator)
pipe.attach(printer)
pipe.attach(kp.io.HDF5Sink, filename=filename)
pipe.drain(10)

####################################################################
# Pandas offers an easy way to read the generated data. As you can
# see, no km3pipe is needed, the only requirement is ``hdf5lib``
# and of course some kind of wrapper library for easy acces, in this case
# Pandas but you can use h5py, pytables directly or other languages.

header = pd.read_hdf(filename, "/header")
event_info = pd.read_hdf(filename, "/event_info")
reco = pd.read_hdf(filename, "/reco")
print(reco.head(3))

####################################################################
# This last section shows how to read the created data with km3pipe
# pipelines.


def inspector(blob):
    reco = blob["Reco"]
    lambdas = reco["lambda"]
    print(lambdas, type(lambdas), lambdas.dtype)
    return blob


pipe = kp.Pipeline()
pipe.attach(kp.io.HDF5Pump, filename=filename)
pipe.attach(printer)
pipe.attach(inspector)
pipe.drain()



#!/usr/bin/env python3


def pandas():
    """Imports and returns ``pandas``."""
    try:
        import pandas
    except ImportError:
        raise ImportError(
            "install the 'pandas' package with:\n\n"
            "    pip install pandas\n\n"
            "or\n\n"
            "    conda install pandas"
        )
    else:
        return pandas


def scipy():
    """Imports and returns ``scipy``."""
    try:
        import scipy
    except ImportError:
        raise ImportError(
            "install the 'scipy' package with:\n\n"
            "    pip install scipy\n\n"
            "or\n\n"
            "    conda install scipy"
        )
    else:
        return scipy


def pypandoc():
    """Imports and returns ``pypandoc``."""
    try:
        import pypandoc
    except ImportError:
        raise ImportError(
            "install the 'pypandoc' package with:\n\n"
            "    pip install pypandoc\n\n"
            "or\n\n"
            "    conda install pypandoc"
        )
    else:
        return pypandoc


def urwid():
    """Imports and returns ``urwid``."""
    try:
        import urwid
    except ImportError:
        raise ImportError(
            "install the 'urwid' package with:\n\n"
            "    pip install urwid\n\n"
            "or\n\n"
            "    conda install urwid"
        )
    else:
        return urwid



# Filename: testing.py
"""
Common unit testing support for km3pipe.

"""
import sys
from functools import wraps
from unittest import TestCase  # noqa
from mock import MagicMock  # noqa
from mock import Mock  # noqa
from mock import patch  # noqa

from numpy.testing import assert_allclose  # noqa
import pytest  # noqa

from km3net_testdata import data_path

skip = pytest.mark.skip
skipif = pytest.mark.skipif

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"



#!/usr/bin/env python
# coding: utf-8 -*-
"""

==================
Calibrating Hits
==================

Hits stored in ROOT and HDF5 files are usually not calibrated, which
means that they have invalid positions, directions and uncorrected hit times.
This example shows how to assign the PMT position and direction to each hit
and applying a time correction to them.

The KM3NeT offline format (derived from aanet) uses a single class for every
hit type (regular hits, MC hits and their correspeonding calibrated and
uncalibrated counter parts). In ROOT files, the actual struct/class definition
is stored along the data, which means that all the attributes are accessible
even when they are invalid or not instantiated. Positions and directions are
also part of these attributes and they are initialised to `(0, 0, 0)`.

The `km3pipe.calib.Calibration()` class can be used to load a calibration and
the `.apply()` method to update the position, direction and correct the arrival
time of hit.

"""

# Author: Tamas Gal <tgal@km3net.de>
# License: BSD-3

import km3pipe as kp
import km3io
from km3net_testdata import data_path

#######################
# The `offline/km3net_offline.root` contains 10 events with hit information:

f = km3io.OfflineReader(data_path("offline/km3net_offline.root"))

#######################
# The corresponding calibration file is stored in `detx/km3net_offline.detx`:

calib = kp.calib.Calibration(filename=data_path("detx/km3net_offline.detx"))

#######################
# Let's grab the hits of the event with index `5`:

hits = f.events[5].hits

#######################
# The positions (pos_x, pos_y, pos_z) and directions (dir_x, dir_y, dir_z) are
# not available for uncalibrated hits, in contrast to aanet where each field
# is present and initialised to some magic value (e.g. 0). km3io hides all
# those fields and do not occupy additional memory for those.

#######################
# Here are the uncalibrated times:

n = 7  # just an arbitrary number to limit the output
uncalibrated_times = hits.t
print(uncalibrated_times[:n])


#######################
# To calibrate the hits, use the `calib.apply()` method which will create a
# `km3pipe.Table`, retrieve the positions and directions of the corresponding
# PMTs, apply the time calibration and also do the PMT time slew correction.

calibrated_hits = calib.apply(hits)

#######################
# The calibrated hits are stored in a `kp.Table` which is a thin wrapper
# around a `numpy.record` array (a simple numpy array with named attributes):

print(calibrated_hits.dtype)


#######################
# The positions and directions are now showing the correct values and the time
# is the calibrated one:

for attr in [f"{k}_{q}" for k in ("pos", "dir") for q in "xzy"]:
    print(attr, getattr(calibrated_hits, attr)[:n])

print(calibrated_hits.time[:n])

#######################
# The `t0` field holds the time calibration correction which was automatically
# added to hit time (`hit.time`):

print(calibrated_hits.t0[:n])


#######################
# As mentioned above, the PMT time slewing correction is also applied, which
# is a tiny correction of the arrival time with respect to the hit's ToT value.
# We can reveal their values by subtracting the t0 from the calibrated time and
# compare to the uncalibrated ones. Notice that hits represented as `kp.Table`
# in km3pipe use `.time` instead of `.t` for mainly historical reasons:

slews = uncalibrated_times - (calibrated_hits.time - calibrated_hits.t0)
print(slews[:n])

#######################
# Let's compare the slews with the ones calculated with `kp.calib.slew()`. The
# values match very well, with tiny variations due to floating point arithmetic:

slew_diff = slews - kp.calib.slew(hits.tot)
print(slew_diff[:n])

######################
# To omit the PMT slew calibration, you can pass the `correct_slewing=False`
# option the `.apply()`:

calibrated_hits_no_slew_correction = calib.apply(hits, correct_slewing=False)

######################
# The difference between the calibration with and without slewing is obviously
# the slewing correction itself:

print((calibrated_hits_no_slew_correction.time - calibrated_hits.time)[:n])



#!/usr/bin/env python3
import tempfile
import unittest

from km3net_testdata import data_path

import km3pipe as kp
import km3modules as km
import numpy as np
import km3io
import awkward as ak


class TestOfflineHeaderTabulator(unittest.TestCase):
    def test_module(self):
        outfile = tempfile.NamedTemporaryFile(delete=True)

        pipe = kp.Pipeline()
        pipe.attach(kp.io.OfflinePump, filename=data_path("offline/numucc.root"))
        pipe.attach(km.io.OfflineHeaderTabulator)
        pipe.attach(kp.io.HDF5Sink, filename=outfile.name)
        pipe.drain()

        pipe = kp.Pipeline()
        pipe.attach(kp.io.HDF5Pump, filename=outfile.name)
        pipe.attach(km.common.Observer, count=10, required_keys=["RawHeader"])
        pipe.drain()


class TestEventInfoTabulator(unittest.TestCase):
    def test_module(self):
        outfile = tempfile.NamedTemporaryFile(delete=True)

        pipe = kp.Pipeline()
        pipe.attach(
            kp.io.OfflinePump,
            filename=data_path(
                "offline/mcv6.0.gsg_muon_highE-CC_50-500GeV.km3sim.jterbr00008357.jorcarec.aanet.905.root"
            ),
        )
        pipe.attach(km.io.EventInfoTabulator)
        pipe.attach(kp.io.HDF5Sink, filename=outfile.name)
        pipe.drain(10)

        pipe = kp.Pipeline()
        pipe.attach(kp.io.HDF5Pump, filename=outfile.name)
        pipe.attach(km.common.Observer, count=10, required_keys=["EventInfo"])
        pipe.attach(CheckW2listContents)
        pipe.drain()


class TestHitsTabulator(unittest.TestCase):
    def test_offline_hits(self):
        outfile = tempfile.NamedTemporaryFile(delete=True)

        pipe = kp.Pipeline()
        pipe.attach(kp.io.OfflinePump, filename=data_path("offline/numucc.root"))
        pipe.attach(km.io.HitsTabulator, kind="offline")
        pipe.attach(kp.io.HDF5Sink, filename=outfile.name)
        pipe.drain()

        pipe = kp.Pipeline()
        pipe.attach(kp.io.HDF5Pump, filename=outfile.name)
        pipe.attach(km.common.Observer, count=10, required_keys=["Hits"])
        pipe.drain()

    def test_mc_hits(self):
        outfile = tempfile.NamedTemporaryFile(delete=True)

        pipe = kp.Pipeline()
        pipe.attach(kp.io.OfflinePump, filename=data_path("offline/numucc.root"))
        pipe.attach(km.io.HitsTabulator, kind="mc")
        pipe.attach(kp.io.HDF5Sink, filename=outfile.name)
        pipe.drain()

        pipe = kp.Pipeline()
        pipe.attach(kp.io.HDF5Pump, filename=outfile.name)
        pipe.attach(km.common.Observer, count=10, required_keys=["McHits"])
        pipe.drain()


class TestMCTracksTabulator(unittest.TestCase):
    def test_module(self):
        outfile = tempfile.NamedTemporaryFile(delete=True)

        pipe = kp.Pipeline()
        pipe.attach(kp.io.OfflinePump, filename=data_path("offline/numucc.root"))
        pipe.attach(km.io.MCTracksTabulator)
        pipe.attach(kp.io.HDF5Sink, filename=outfile.name)
        pipe.drain()

        pipe = kp.Pipeline()
        pipe.attach(kp.io.HDF5Pump, filename=outfile.name)
        pipe.attach(km.common.Observer, count=10, required_keys=["McTracks"])
        pipe.drain()


class TestRecoTracksTabulator(unittest.TestCase):
    def test_module(self):
        outfile = tempfile.NamedTemporaryFile(delete=True)

        pipe = kp.Pipeline()
        pipe.attach(
            kp.io.OfflinePump,
            filename=data_path(
                "offline/mcv6.0.gsg_muon_highE-CC_50-500GeV.km3sim.jterbr00008357.jorcarec.aanet.905.root"
            ),
        )
        pipe.attach(km.io.RecoTracksTabulator, best_tracks=True)
        pipe.attach(kp.io.HDF5Sink, filename=outfile.name)
        pipe.drain(5)

        pipe = kp.Pipeline()
        pipe.attach(kp.io.HDF5Pump, filename=outfile.name)
        pipe.attach(km.common.Observer, count=5, required_keys=["Tracks"])
        pipe.attach(km.common.Observer, count=5, required_keys=["RecStages"])
        pipe.attach(km.common.Observer, count=5, required_keys=["BestJmuon"])
        pipe.attach(CheckRecoContents)
        pipe.drain()


class CheckRecoContents(kp.Module):
    def configure(self):

        # use this to count through the single events
        self.event_idx = 0

        # get the original file to compare to
        filename = data_path(
            "offline/mcv6.0.gsg_muon_highE-CC_50-500GeV.km3sim.jterbr00008357.jorcarec.aanet.905.root"
        )
        self.f = km3io.OfflineReader(filename)

    def process(self, blob):

        # first, get some extracted values form the h5 file

        # best track
        jmuon_dir_z = blob["BestJmuon"].dir_z  # a reco parameter
        jmuon_jgandalf_chi2 = blob["BestJmuon"].JGANDALF_CHI2  # a fitinf parameter

        # all tracks
        tracks_dir_z = blob["Tracks"].dir_z
        tracks_jgandalf_chi2 = blob["Tracks"].JGANDALF_CHI2

        # then, get the values from the original file

        # all tracks
        all_tracks_raw = self.f.events[self.event_idx].tracks
        all_tracks_raw_dir_z = all_tracks_raw.dir_z
        all_tracks_raw_fitinf = self._preprocess_fitinf(all_tracks_raw.fitinf)
        all_tracks_raw_jgandalf_chi2 = all_tracks_raw_fitinf[
            :, km3io.definitions.fitparameters.JGANDALF_CHI2
        ]

        # best tracks
        best_tracks_raw = km3io.tools.best_jmuon(all_tracks_raw)
        best_tracks_raw_dir_z = best_tracks_raw.dir_z
        best_tracks_raw_jgandalf_chi2 = best_tracks_raw.fitinf[
            km3io.definitions.fitparameters.JGANDALF_CHI2
        ]

        # and finally compare
        assert np.allclose(best_tracks_raw_dir_z, jmuon_dir_z)
        assert np.allclose(best_tracks_raw_jgandalf_chi2, jmuon_jgandalf_chi2)

        # since an ak array wise assertion is not possible, do this element wise
        for i in range(len(all_tracks_raw_dir_z)):

            assert np.allclose(all_tracks_raw_dir_z[i], tracks_dir_z[i])

            # exclude nans from the assertion as they are not detected as equal
            if not np.isnan(all_tracks_raw_jgandalf_chi2[i]) and not np.isnan(
                tracks_jgandalf_chi2[i]
            ):
                assert np.allclose(
                    all_tracks_raw_jgandalf_chi2[i], tracks_jgandalf_chi2[i]
                )

        self.event_idx += 1

        return blob

    def _preprocess_fitinf(self, fitinf):
        # preprocess the fitinf a bit - yay!
        n_columns = max(km3io.definitions.fitparameters.values()) + 1
        fitinf_array = np.ma.filled(
            ak.to_numpy(ak.pad_none(fitinf, target=n_columns, axis=-1)),
            fill_value=np.nan,
        ).astype("float32")
        return fitinf_array


class CheckW2listContents(kp.Module):
    def configure(self):

        # use this to count through the single events
        self.event_idx = 0

        # get the original file to compare to
        filename = data_path(
            "offline/mcv6.0.gsg_muon_highE-CC_50-500GeV.km3sim.jterbr00008357.jorcarec.aanet.905.root"
        )
        self.f = km3io.OfflineReader(filename)

    def process(self, blob):

        # extracted values
        by = blob["EventInfo"].W2LIST_GSEAGEN_BY[0]
        cc = blob["EventInfo"].W2LIST_GSEAGEN_CC[0]

        # original values
        original_by = self.f.events[self.event_idx].w2list[
            km3io.definitions.w2list_gseagen["W2LIST_GSEAGEN_BY"]
        ]
        original_cc = self.f.events[self.event_idx].w2list[
            km3io.definitions.w2list_gseagen["W2LIST_GSEAGEN_CC"]
        ]

        # and compare
        assert by == original_by
        assert cc == original_cc

        self.event_idx += 1

        return blob



#!/usr/bin/env python3
# Filename: h5extract.py
"""
A tool to extract data from KM3NeT ROOT files to HDF5.

Usage:
    h5extract [options] FILENAME
    h5extract (-h | --help)
    h5extract --version

Options:
    -o OUTFILE                  Output file.
    -n N_EVENTS                 Number of events to extract.
    --offline-header            The header of an offline file.
    --event-info                General event information.
    --offline-hits              Offline hits.
    --with-calibration          Include calibration information for offline hits.
    --mc-hits                   MC hits (use with care!).
    --online-hits               Snapshot and triggered hits (combined).
    --mc-tracks                 MC tracks..
    --mc-tracks-usr-data        "usr" data from MC tracks (this will be slow).
    --reco-tracks               Reconstructed tracks.
    --best-tracks               Write a separate best track dataset for each known reco.
    --provenance-file=FILENAME  The file to store the provenance information.
    --timeit                    Print detailed pipeline performance statistics.
    --step-size=N               Number of events to cache or amount of data [default: 2000].
    --aashower-legacy           Use only with old MC files where aashower rec_stage=100 (now it is 300)
    -h --help                   Show this screen.
    --version                   Show the version.

"""
from thepipe import Provenance
import km3pipe as kp
import km3modules as km


def main():
    from docopt import docopt

    args = docopt(__doc__, version=kp.version)
    step_size = int(args["--step-size"])

    default_flags = (
        "--offline-header",
        "--event-info",
        "--offline-hits",
        "--mc-hits",
        "--mc-tracks",
        "--with-calibration",
        "--reco-tracks",
        "--best-tracks",
    )
    if not any([args[k] for k in default_flags]):
        for k in default_flags:
            args[k] = True

    outfile = args["-o"]
    if outfile is None:
        outfile = args["FILENAME"] + ".h5"

    provfile = args["--provenance-file"]
    if provfile is None:
        provfile = outfile + ".prov.json"

    Provenance().outfile = provfile
    pipe = kp.Pipeline(timeit=args["--timeit"])
    pipe.attach(kp.io.OfflinePump, filename=args["FILENAME"], step_size=step_size)
    pipe.attach(km.StatusBar, every=1000)
    pipe.attach(km.common.MemoryObserver, every=500)
    if args["--offline-header"]:
        pipe.attach(km.io.OfflineHeaderTabulator)
    if args["--event-info"]:
        pipe.attach(km.io.EventInfoTabulator)
    if args["--offline-hits"]:
        pipe.attach(
            km.io.HitsTabulator,
            name="Offline",
            kind="offline",
            with_calibration=args["--with-calibration"],
        )
    if args["--online-hits"]:
        pipe.attach(km.io.HitsTabulator, name="Online", kind="online")
    if args["--mc-hits"]:
        pipe.attach(km.io.HitsTabulator, name="MC", kind="mc")
    if args["--mc-tracks"]:
        pipe.attach(km.io.MCTracksTabulator, read_usr_data=args["--mc-tracks-usr-data"])
    if args["--reco-tracks"]:
        pipe.attach(
            km.io.RecoTracksTabulator,
            best_tracks=args["--best-tracks"],
            aashower_legacy=args["--aashower-legacy"],
        )
    pipe.attach(kp.io.HDF5Sink, filename=outfile)
    if args["-n"] is not None:
        pipe.drain(int(args["-n"]))
    else:
        pipe.drain()



# Filename: test_physics.py
# pylint: disable=locally-disabled

import km3io as ki
import numpy as np
import awkward as ak

from km3pipe.dataclasses import Table
from km3pipe.testing import TestCase, data_path
from km3pipe.hardware import Detector
from km3pipe.calib import Calibration
from km3pipe.physics import cherenkov, get_closest, cut4d
import km3pipe.extras


__author__ = "Zineb ALY"
__copyright__ = "Copyright 2017, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Zineb ALY"
__email__ = "zaly@km3net.de"
__status__ = "Development"


class TestGetCherenkov(TestCase):
    def setUp(self):
        self.calib_hits = {
            "pos_x": [
                19.522,
                19.435,
                19.496,
                19.544,
                9.48,
                9.644,
                9.441,
                9.739,
                9.714,
                9.72,
            ],
            "pos_y": [
                -12.053,
                -12.194,
                -12.196,
                -12.289,
                6.013,
                6.086,
                5.852,
                5.769,
                6.02,
                5.786,
            ],
            "pos_z": [
                76.662,
                76.662,
                76.381,
                76.381,
                116.249,
                116.079,
                116.249,
                66.762,
                66.71,
                66.932,
            ],
            "dir_x": [
                -0.39,
                -0.831,
                -0.527,
                -0.279,
                -0.606,
                0.22,
                -0.797,
                0.696,
                0.57,
                0.606,
            ],
            "dir_y": [
                0.735,
                0.03,
                0.018,
                -0.447,
                0.57,
                0.93,
                -0.239,
                -0.655,
                0.606,
                -0.57,
            ],
            "dir_z": [
                0.555,
                0.555,
                -0.85,
                -0.85,
                0.555,
                -0.295,
                0.555,
                -0.295,
                -0.555,
                0.555,
            ],
        }

        self.track = {
            "pos_x": 1.7524502152598151,
            "pos_y": 39.06202405657308,
            "pos_z": 130.44049806891948,
            "dir_x": 0.028617421257374293,
            "dir_y": -0.489704257367248,
            "dir_z": -0.8714188335794505,
            "t": 70311441.452294,
        }

    def test_cherenkov_from_dict(self):

        arr = cherenkov(self.calib_hits, self.track)

        self.assertAlmostEqual(arr["d_photon_closest"][0], 24.049593557846112)
        self.assertAlmostEqual(arr["d_photon_closest"][1], 24.085065395206847)

        self.assertAlmostEqual(arr["d_photon"][0], 35.80244420413484)
        self.assertAlmostEqual(arr["d_photon"][1], 35.855250854478896)

        self.assertAlmostEqual(arr["d_track"][0], 45.88106599210481)
        self.assertAlmostEqual(arr["d_track"][1], 45.90850564175342)

        self.assertAlmostEqual(arr["t_photon"][0], 70311759.26448613)
        self.assertAlmostEqual(arr["t_photon"][1], 70311759.59904088)

        self.assertAlmostEqual(arr["cos_photon_PMT"][0], -0.98123942583677)
        self.assertAlmostEqual(arr["cos_photon_PMT"][1], -0.6166369315726149)

        self.assertAlmostEqual(arr["dir_x_photon"][0], 0.45964884122649263)
        self.assertAlmostEqual(arr["dir_x_photon"][1], 0.45652355929477095)

        self.assertAlmostEqual(arr["dir_y_photon"][0], -0.8001372907490844)
        self.assertAlmostEqual(arr["dir_y_photon"][1], -0.8025165828910586)

        self.assertAlmostEqual(arr["dir_z_photon"][0], -0.3853612055096594)
        self.assertAlmostEqual(arr["dir_z_photon"][1], -0.38412676812960095)

    def test_cherenkov_from_Table(self):

        arr = cherenkov(Table(self.calib_hits), Table(self.track))

        self.assertAlmostEqual(arr["d_photon_closest"][0], 24.049593557846112)
        self.assertAlmostEqual(arr["d_photon"][0], 35.80244420413484)
        self.assertAlmostEqual(arr["d_track"][0], 45.88106599210481)
        self.assertAlmostEqual(arr["t_photon"][0], 70311759.26448613)
        self.assertAlmostEqual(arr["cos_photon_PMT"][0], -0.98123942583677)
        self.assertAlmostEqual(arr["dir_x_photon"][0], 0.45964884122649263)
        self.assertAlmostEqual(arr["dir_y_photon"][0], -0.8001372907490844)
        self.assertAlmostEqual(arr["dir_z_photon"][0], -0.3853612055096594)

    def test_cherenkov_from_DataFrame(self):

        pd = km3pipe.extras.pandas()

        arr = cherenkov(pd.DataFrame(self.calib_hits), pd.Series(self.track))

        self.assertAlmostEqual(arr["d_photon_closest"][0], 24.049593557846112)
        self.assertAlmostEqual(arr["d_photon"][0], 35.80244420413484)
        self.assertAlmostEqual(arr["d_track"][0], 45.88106599210481)
        self.assertAlmostEqual(arr["t_photon"][0], 70311759.26448613)
        self.assertAlmostEqual(arr["cos_photon_PMT"][0], -0.98123942583677)
        self.assertAlmostEqual(arr["dir_x_photon"][0], 0.45964884122649263)
        self.assertAlmostEqual(arr["dir_y_photon"][0], -0.8001372907490844)
        self.assertAlmostEqual(arr["dir_z_photon"][0], -0.3853612055096594)

    def test_cherenkov_from_best_track_which_returns_awkward_Record(self):
        pd = km3pipe.extras.pandas()

        best_track = ak.Record(self.track)

        arr = cherenkov(pd.DataFrame(self.calib_hits), best_track)

        self.assertAlmostEqual(arr["d_photon_closest"][0], 24.049593557846112)
        self.assertAlmostEqual(arr["d_photon"][0], 35.80244420413484)
        self.assertAlmostEqual(arr["d_track"][0], 45.88106599210481)
        self.assertAlmostEqual(arr["t_photon"][0], 70311759.26448613)
        self.assertAlmostEqual(arr["cos_photon_PMT"][0], -0.98123942583677)
        self.assertAlmostEqual(arr["dir_x_photon"][0], 0.45964884122649263)
        self.assertAlmostEqual(arr["dir_y_photon"][0], -0.8001372907490844)
        self.assertAlmostEqual(arr["dir_z_photon"][0], -0.3853612055096594)


class TestGetClosest(TestCase):
    def setUp(self):
        self.track = {
            "pos_x": 1.7524502152598151,
            "pos_y": 39.06202405657308,
            "pos_z": 130.44049806891948,
            "dir_x": 0.028617421257374293,
            "dir_y": -0.489704257367248,
            "dir_z": -0.8714188335794505,
        }

        self.det = Detector(data_path("detx/detx_v3.detx"))

        pd = km3pipe.extras.pandas()

        self.DU = pd.DataFrame(self.det.dom_table).mean()

    def test_get_closest(self):
        pd = km3pipe.extras.pandas()

        DU = pd.DataFrame(self.det.dom_table).mean()
        d_closest, z_closest = get_closest(self.track, DU)

        self.assertAlmostEqual(d_closest, 9.073491762564467)
        self.assertAlmostEqual(z_closest, 82.24928115091757)

    def test_get_closest_from_DataFrame(self):
        pd = km3pipe.extras.pandas()
        d_closest, z_closest = get_closest(pd.Series(self.track), pd.Series(self.DU))

        self.assertAlmostEqual(d_closest, 9.073491762564467)
        self.assertAlmostEqual(z_closest, 82.24928115091757)

    def test_get_closest_from_Table(self):
        trk = {
            "pos_x": 1.7524502152598151,
            "pos_y": 39.06202405657308,
            "pos_z": 130.44049806891948,
            "dir_x": 0.028617421257374293,
            "dir_y": -0.489704257367248,
            "dir_z": -0.8714188335794505,
        }

        pd = km3pipe.extras.pandas()
        mean = pd.DataFrame(self.det.dom_table).mean()

        d_closest, z_closest = get_closest(Table(trk), mean)

        self.assertAlmostEqual(d_closest, 9.073491762564467)
        self.assertAlmostEqual(z_closest, 82.24928115091757)


class TestCut4D(TestCase):
    def test_cut4d(self):
        point4d = Table({"pos_x": [0], "pos_y": [3], "pos_z": [0], "t": [20]})

        items = Table(
            {
                "pos_x": [0, 10, 0, 20, 0],
                "pos_y": [10, 0, 0, 0, 30],
                "pos_z": [0, 0, 10, 0, 0],
                "time": [60, 15, 40, 20, 100],
            }
        )

        tmin = -50.0
        tmax = 10.0
        rmin = 3.0
        rmax = 80.0

        selected_items = cut4d(point4d, tmin, tmax, rmin, rmax, items)
        assert len(selected_items) == 3

        self.assertListEqual(
            list(np.array([0, 10, 0, 60])),
            list(selected_items.T[0]),
        )
        self.assertListEqual(
            list(np.array([0, 0, 10, 40])),
            list(selected_items.T[1]),
        )
        self.assertListEqual(
            list(np.array([0, 30, 0, 100])),
            list(selected_items.T[2]),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
=========================
K40 Intra-DOM Calibration
=========================

The following script calculates the PMT time offsets using K40 coincidences

"""

# Author: Jonas Reubelt <jreubelt@km3net.de> and Tamas Gal <tgal@km3net.de>
# License: MIT
import km3pipe as kp
from km3modules import k40
from km3modules.common import StatusBar, MemoryObserver, Siphon
from km3modules.plot import IntraDOMCalibrationPlotter
import km3pipe.style

km3pipe.style.use("km3pipe")

pipe = kp.Pipeline(timeit=True)
pipe.attach(
    kp.io.ch.CHPump,
    host="127.0.0.1",
    port=5553,
    tags="IO_TSL, IO_MONIT",
    timeout=7 * 60 * 60 * 24,
    max_queue=42,
)
pipe.attach(kp.io.ch.CHTagger)
pipe.attach(StatusBar, every=1000)
pipe.attach(MemoryObserver, every=5000)
pipe.attach(k40.MedianPMTRatesService, only_if="IO_MONIT")
pipe.attach(kp.io.daq.TimesliceParser)
pipe.attach(k40.TwofoldCounter, tmax=10)
pipe.attach(Siphon, volume=10 * 10 * 1, flush=True)
pipe.attach(k40.K40BackgroundSubtractor)
pipe.attach(k40.IntraDOMCalibrator, ctmin=0.0)
pipe.attach(IntraDOMCalibrationPlotter)
pipe.attach(k40.ResetTwofoldCounts)
pipe.drain()



# Filename: constants.py
# pylint: disable=C0103
# pragma: no cover
"""
The constants used in KM3Pipe.

"""
import math

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

# Physics constants used in physics module
WATER_INDEX = 1.3499  # Used in aanet
INDEX_OF_REFRACTION_WATER = 1.3800851282  # Used in Jpp (e.g. in PDFs)
DN_DL = 0.0298
COS_CHERENKOV = 1 / WATER_INDEX
CHERENKOV_ANGLE_RAD = math.acos(COS_CHERENKOV)
SIN_CHERENKOV = math.sin(CHERENKOV_ANGLE_RAD)
TAN_CHERENKOV = math.tan(CHERENKOV_ANGLE_RAD)
C_LIGHT = 299792458e-9  # m/ns
V_LIGHT_WATER = C_LIGHT / (WATER_INDEX + DN_DL)
C_WATER = C_LIGHT / INDEX_OF_REFRACTION_WATER

c = 2.99792458e8  # m/s



# Filename: test_time.py
# pylint: disable=locally-disabled,C0111,R0904,C0103

import sqlite3
import tempfile

import km3pipe as kp
from km3pipe.dataclasses import Table
from km3modules.common import (
    Siphon,
    Delete,
    Keep,
    Dump,
    StatusBar,
    TickTock,
    MemoryObserver,
    BlobIndexer,
    LocalDBService,
    Observer,
    MultiFilePump,
    FilePump,
)
from km3pipe.testing import TestCase, MagicMock
from km3pipe.tools import istype

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class InfinitePump(kp.Module):
    """A pump which just infinetly spits out indexed blobs"""

    def configure(self):
        self.i = 0

    def process(self, blob):
        self.i += 1
        blob["i"] = self.i
        return blob


class TestKeep(TestCase):
    def test_keep_a_single_key(self):
        class APump(kp.Module):
            def process(self, blob):
                blob["a"] = "a"
                blob["b"] = "b"
                blob["c"] = "c"
                blob["d"] = "d"
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert "a" not in blob
                assert "b" not in blob
                assert "c" not in blob
                assert "d" == blob["d"]
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, keys="d")
        pipe.attach(Observer)
        pipe.drain(5)

    def test_keep_multiple_keys(self):
        class APump(kp.Module):
            def process(self, blob):
                blob["a"] = "a"
                blob["b"] = "b"
                blob["c"] = "c"
                blob["d"] = "d"
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert "a" not in blob
                assert "b" == blob["b"]
                assert "c" not in blob
                assert "d" == blob["d"]
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, keys=["b", "d"])
        pipe.attach(Observer)
        pipe.drain(5)

    def test_hdf5_keep_group_wo_subgroup(self):
        class APump(kp.Module):
            def process(self, blob):
                blob["A"] = kp.Table(
                    {"foo": [1, 2, 3], "bar": [4, 5, 6]}, h5loc="/foobar"
                )
                blob["B"] = kp.Table(
                    {"a": [1.1, 2.1, 3.1], "b": [4.2, 5.2, 6.2]}, h5loc="/ab"
                )
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert "A" in blob.keys()
                assert "/foobar" == blob["A"].h5loc
                assert not "B" in blob.keys()
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, h5locs=["/foobar"])
        pipe.attach(Observer)
        pipe.drain(5)

    def test_hdf5_keep_group_w_subgroup(self):
        class APump(kp.Module):
            def process(self, blob):
                blob["A"] = kp.Table(
                    {"foo": [1, 2, 3], "bar": [4, 5, 6]}, h5loc="/foobar"
                )
                blob["B"] = kp.Table(
                    {"a": [1.1, 2.1, 3.1], "b": [4.2, 5.2, 6.2]}, h5loc="/ab"
                )
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert "A" in blob.keys()
                assert "/foobar" == blob["A"].h5loc
                assert not "B" in blob.keys()
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, h5locs=["/foobar"])
        pipe.attach(Observer)
        pipe.drain(5)

    def test_key_hdf5_group_individual(self):
        class APump(kp.Module):
            def process(self, blob):
                blob["A"] = kp.Table(
                    {"foo": [1, 2, 3], "bar": [4, 5, 6]}, h5loc="/foobar"
                )
                blob["B"] = kp.Table(
                    {"a": [1.1, 2.1, 3.1], "b": [4.2, 5.2, 6.2]}, h5loc="/ab"
                )
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert "A" in blob.keys()
                assert "/foobar" == blob["A"].h5loc
                assert "B" in blob.keys()
                assert "/ab" == blob["B"].h5loc
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, keys=["B"], h5locs=["/foobar"])
        pipe.attach(Observer)
        pipe.drain(5)

    def test_key_hdf5_group_parallel(self):
        class APump(kp.Module):
            def process(self, blob):
                blob["A"] = kp.Table(
                    {"foo": [1, 2, 3], "bar": [4, 5, 6]}, h5loc="/foobar"
                )
                blob["B"] = kp.Table(
                    {"a": [1.1, 2.1, 3.1], "b": [4.2, 5.2, 6.2]}, h5loc="/ab"
                )
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert "A" in blob.keys()
                assert "/foobar" == blob["A"].h5loc
                assert not "B" in blob.keys()
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, keys=["A"], h5locs=["/foobar"])
        pipe.attach(Observer)
        pipe.drain(5)

    def test_major_hdf5_group(self):
        class APump(kp.Module):
            def process(self, blob):
                blob["A"] = kp.Table(
                    {"foo": [1, 2, 3], "bar": [4, 5, 6]}, h5loc="/foobar/a"
                )
                blob["B"] = kp.Table(
                    {"a": [1.1, 2.1, 3.1], "b": [4.2, 5.2, 6.2]}, h5loc="/foobar/b"
                )
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert "A" in blob.keys()
                assert "/foobar/a" == blob["A"].h5loc
                assert "B" in blob.keys()
                assert "/foobar/b" == blob["B"].h5loc
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, h5locs=["/foobar"])
        pipe.attach(Observer)
        pipe.drain(5)

    def test_major_hdf5_group_nested(self):
        class APump(kp.Module):
            def process(self, blob):
                blob["A"] = kp.Table({"a": 0}, h5loc="/foo/bar/a")
                blob["B"] = kp.Table({"b": 1}, h5loc="/foo/bar/baz/b")
                blob["C"] = kp.Table({"c": 2}, h5loc="/foo/bar/baz/fjord/c")
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert "A" not in blob
                assert "B" in blob
                assert "C" in blob
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, h5locs=["/foo/bar/baz"])
        pipe.attach(Observer)
        pipe.drain(5)


class TestDelete(TestCase):
    def test_delete_a_single_key(self):
        class APump(kp.Module):
            def process(self, blob):
                blob["a"] = "a"
                blob["b"] = "b"
                blob["c"] = "c"
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert "a" == blob["a"]
                assert "b" not in blob
                assert "c" == blob["c"]
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Delete, key="b")
        pipe.attach(Observer)
        pipe.drain(5)

    def test_delete_multiple_keys(self):
        class APump(kp.Module):
            def process(self, blob):
                blob["a"] = "a"
                blob["b"] = "b"
                blob["c"] = "c"
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert "a" not in blob
                assert "b" not in blob
                assert "c" == blob["c"]
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Delete, keys=["a", "b"])
        pipe.attach(Observer)
        pipe.drain(5)


class TestSiphon(TestCase):
    def test_siphon(self):
        class Observer(kp.Module):
            def configure(self):
                self.mock = MagicMock()

            def process(self, blob):
                self.mock()
                return blob

            def finish(self):
                assert self.mock.call_count == 7

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(Siphon, volume=10)
        pipe.attach(Observer)
        pipe.drain(17)

    def test_siphon_with_flush(self):
        class Observer(kp.Module):
            def configure(self):
                self.mock = MagicMock()

            def process(self, blob):
                self.mock()
                return blob

            def finish(self):
                assert self.mock.call_count == 1

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(Siphon, volume=10, flush=True)
        pipe.attach(Observer)
        pipe.drain(21)

    def test_siphon_with_flush_2(self):
        class Observer(kp.Module):
            def configure(self):
                self.mock = MagicMock()

            def process(self, blob):
                self.mock()
                return blob

            def finish(self):
                assert self.mock.call_count == 2

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(Siphon, volume=10, flush=True)
        pipe.attach(Observer)
        pipe.drain(22)


class TestDump(TestCase):
    def test_dump(self):
        def add_something(blob):
            blob["a"] = 1
            return blob

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(add_something)
        pipe.attach(Dump)
        pipe.drain(3)

    def test_dump_a_key(self):
        def add_something(blob):
            blob["a"] = 1
            return blob

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(add_something)
        pipe.attach(Dump, key="a")
        pipe.drain(3)

    def test_dump_multiple_keys(self):
        def add_something(blob):
            blob["a"] = 1
            blob["b"] = 2
            return blob

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(add_something)
        pipe.attach(Dump, keys=["a", "b"])
        pipe.drain(3)

    def test_dump_full(self):
        def add_something(blob):
            blob["a"] = 1
            blob["b"] = 2
            return blob

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(add_something)
        pipe.attach(Dump, full=True)
        pipe.drain(3)


class TestStatusbar(TestCase):
    def test_statusbar(self):
        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(StatusBar, every=2)
        pipe.drain(5)


class TestTickTock(TestCase):
    def test_ticktock(self):
        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(TickTock)
        pipe.drain(5)


class TestMemoryObserver(TestCase):
    def test_memory_observer(self):
        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(MemoryObserver)
        pipe.drain(5)


class TestBlobIndexer(TestCase):
    def test_blob_indexer(self):
        class Observer(kp.Module):
            def configure(self):
                self.index = 0

            def process(self, blob):
                assert blob["blob_index"] == self.index
                self.index += 1
                return blob

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(BlobIndexer)
        pipe.attach(Observer)
        pipe.drain(4)


class TestLocalDBService(TestCase):
    def test_create_table(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        dbs = LocalDBService(filename=fobj.name)
        dbs.create_table("foo", ["a", "b"], ["INT", "TEXT"])
        assert dbs.table_exists("foo")

    def test_create_table_does_not_overwrite_by_default(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        dbs = LocalDBService(filename=fobj.name)
        dbs.create_table("foo", ["a", "b"], ["INT", "TEXT"])
        with self.assertRaises(sqlite3.OperationalError):
            dbs.create_table("foo", ["a", "b"], ["INT", "TEXT"])

    def test_create_table_allows_overwrite(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        dbs = LocalDBService(filename=fobj.name)
        dbs.create_table("foo", ["a", "b"], ["INT", "TEXT"])
        dbs.create_table("foo", ["a", "b"], ["INT", "TEXT"], overwrite=True)

    def test_insert_row(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        dbs = LocalDBService(filename=fobj.name)
        dbs.create_table("foo", ["a", "b"], ["INT", "TEXT"])

        dbs.insert_row("foo", ["a", "b"], (23, "42"))
        dbs.insert_row("foo", ["a", "b"], (5, "hello"))

        cur = dbs.connection.cursor()
        cur.execute("SELECT * FROM foo")
        data = cur.fetchall()
        assert 2 == len(data)
        assert 23 == data[0][0]
        assert "42" == data[0][1]
        assert 5 == data[1][0]
        assert "hello" == data[1][1]


class TestObserver(TestCase):
    def test_observer(self):
        class Dummy(kp.Module):
            def process(self, blob):
                blob["a"] = 1
                return blob

        pipe = kp.Pipeline()
        pipe.attach(Dummy)
        pipe.attach(Observer, count=5, required_keys="a")
        pipe.drain(5)

    def test_observer_raises_when_count_wrong(self):
        class Dummy(kp.Module):
            def process(self, blob):
                return blob

        pipe = kp.Pipeline()
        pipe.attach(Dummy)
        pipe.attach(Observer, count=5)

        with self.assertRaises(AssertionError):
            pipe.drain(2)

    def test_observer_raises_when_key_is_missing(self):
        class Dummy(kp.Module):
            def process(self, blob):
                blob["a"] = 1
                return blob

        pipe = kp.Pipeline()
        pipe.attach(Dummy)
        pipe.attach(Observer, required_keys=["b"])

        with self.assertRaises(AssertionError):
            pipe.drain(2)


class TestMultiFilePump(TestCase):
    def test_iteration(self):
        class DummyPump(kp.Module):
            def configure(self):
                self.idx = 0
                self.max_iterations = self.get("max_iterations", default=5)
                self.blobs = self.blob_generator()
                assert 23 == self.get("foo")
                assert "narf" == self.get("bar")

            def process(self, blob):
                return next(self)

            def blob_generator(self):
                for idx in range(self.max_iterations):
                    yield kp.Blob({"index": self.idx, "tab": kp.Table({"a": 1})})

            def finish(self):
                return self.idx

            def __iter__(self):
                return self

            def __next__(self):
                return next(self.blobs)

        filenames = ["a", "b", "c"]
        max_iterations = 5
        total_iterations = max_iterations * len(filenames)

        super_self = self

        class Observer(kp.Module):
            def configure(self):
                self.count = 0
                self.filenames = []
                self.group_id = []

            def process(self, blob):
                self.count += 1
                self.filenames.append(blob["filename"])
                self.group_id.append(blob["tab"].group_id[0])
                return blob

            def finish(self):
                assert self.count == total_iterations
                assert "".join(f * max_iterations for f in filenames) == "".join(
                    self.filenames
                )
                super_self.assertListEqual(list(range(total_iterations)), self.group_id)

        pipe = kp.Pipeline()
        pipe.attach(
            MultiFilePump,
            pump=DummyPump,
            filenames=filenames,
            max_iterations=max_iterations,
            kwargs={"foo": 23, "bar": "narf"},
        )
        pipe.attach(Observer)
        pipe.drain()


class TestFilePump(TestCase):
    def test_iteration(self):
        filenames = ["a", "b", "c"]

        super_self = self

        class Observer(kp.Module):
            def configure(self):
                self.count = 0
                self.filenames = []

            def process(self, blob):
                self.count += 1
                self.filenames.append(blob["filename"])
                return blob

            def finish(self):
                assert self.count == len(filenames)
                super_self.assertListEqual(filenames, self.filenames)

        pipe = kp.Pipeline()
        pipe.attach(FilePump, filenames=filenames)
        pipe.attach(Observer)
        pipe.drain()
